# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from typing import Dict, List, Literal, Optional

from tqdm import tqdm
from wakepy import keep

from ansys.simai.core.data.base import ComputableDataModel, Directory
from ansys.simai.core.data.geometries import Geometry
from ansys.simai.core.data.types import (
    Identifiable,
    get_id_from_identifiable,
    get_object_from_identifiable,
)
from ansys.simai.core.errors import InvalidArguments

logger = logging.getLogger(__name__)


class _OptimizationTrialRun(ComputableDataModel):
    """Provides the local representation of an optimization trial run object.

    The optimization trial run is an iteration of the optimization process.
    Each trial run tests a geometry and returns new parameters for the next geometry to try.
    """


# Undocumented for now, users don't really need to interact with it
class _OptimizationTrialRunDirectory(Directory[_OptimizationTrialRun]):
    _data_model = _OptimizationTrialRun

    def get(self, trial_run_id: str):
        """Get a specific trial run from the server."""
        return self._model_from(self._client._api.get_optimization_trial_run(trial_run_id))

    def _run_iteration(
        self, optimization: Identifiable["Optimization"], parameters: Dict
    ) -> _OptimizationTrialRun:
        optimization_id = get_id_from_identifiable(optimization)
        return self._model_from(
            self._client._api.run_optimization_trial(optimization_id, parameters)
        )


class Optimization(ComputableDataModel):
    """Provides the local representation of an optimization definition object."""

    def _run_iteration(self, parameters: Dict) -> "_OptimizationTrialRun":
        return self._client._optimization_trial_run_directory._run_iteration(self.id, parameters)


class OptimizationResult:
    """Result for a non-parametric optimization."""

    def __init__(
        self,
        optimization: "Optimization",
        objectives: list[Dict],
        geometries: list[Dict],
    ):
        self._optimization = optimization
        self._objectives: list[Dict] = objectives
        self._geometries: list[Geometry] = [
            self._optimization._client.geometries._model_from(geom) for geom in geometries
        ]

    @property
    def optimization(self) -> Optimization:
        """Optimization object of the optimization run."""
        return self._optimization

    def list_objectives(self) -> List[Dict]:
        """List of objectives generated by the optimization."""
        return self._objectives

    def list_geometries(self) -> List[Geometry]:
        """List of geometries generated by the optimization."""
        return self._geometries


class OptimizationDirectory(Directory[Optimization]):
    """Provides a collection of methods related to optimizations.

    This class is accessed through ``client.optimizations``.

    Example:
        .. code-block:: python

            import ansys.simai.core as asc

            simai_client = asc.from_config()
            simai_client.optimizations.run_non_parametric(...)
    """

    _data_model = Optimization

    def get(self, optimization_id: str) -> Optimization:
        """Get a specific optimization object from the server.

        Args:
            optimization_id: ID of the optimization.

        Returns:
            :py:class:`Optimization`.
        """
        return self._model_from(self._client._api.get_optimization(optimization_id))

    def run_non_parametric(
        self,
        geometry: Identifiable[Geometry],
        bounding_boxes: List[List[float]],
        n_iters: int,
        symmetries: Optional[List[Literal["x", "y", "z", "X", "Y", "Z"]]] = None,
        axial_symmetry: Optional[Literal["x", "y", "z"]] = None,
        scalars: Optional[Dict[str, float]] = None,
        minimize: Optional[List[str]] = None,
        maximize: Optional[List[str]] = None,
        max_displacement: Optional[List[float]] = None,
        show_progress: bool = False,
        boundary_conditions: Optional[Dict[str, float]] = None,
    ) -> OptimizationResult:
        """Run an optimization loop to generate geometries, server-side, using automorphing.
        Automorphing is a non-parametric deformation of a surface geometry.

        Args:
            geometry: Required. The object (Geometry) or the ID (str) of the baseline geometry on which to perform
                the non-parametric optimization and that has already been used to build an AI model.
                The optimization will run in the same workspace as the model.
            bounding_boxes: Required. The list of the bounds of the different boxes that will define the absolute locations
                of the geometry to optimize. It is a list of lists, and each sub-list must have exactly six items with the following
                expected order: ``[x_min, x_max, y_min, y_max, z_min, z_max]``.

                Example format::

                    [
                    [box1_xmin, box1_xmax, box1_ymin, box1_ymax, box1_zmin, box1_zmax],
                    [box2_xmin, box2_xmax, box2_ymin, box2_ymax, box2_zmin, box2_zmax],
                    ...
                    ]

            n_iters: Required. The number of optimization iterations. This number must be a strictly positive integer.
                It will define the number of automorphed geometries uploaded to the SimAI workspace.
            symmetries: Optional. The list of symmetry axes, axes being x, y, and z, defining a plane around which the geometry is mirrored.

                - The planar symmetry is applied to all the ``bounding_boxes`` defined.
                - ``symmetries`` and ``axial_symmetry`` are mutually exclusive parameters.

            axial_symmetry: Optional. The axis, defined by a scalar, along which axial symmetry is applied.
                For the deformation to be axially symmetrical:

                - The center of the bounding box must coincide with the targeted central axis.
                - The mesh should be axially symmetrical in the bounding box.
                - The ``axial_symmetry`` is applied to all the ``bounding_boxes`` defined.
                - ``symmetries`` and ``axial_symmetry`` are mutually exclusive parameters.

            scalars: Optional. The values of the scalars to perform the optimization at.
                The values must correspond to existing scalars already defined in your SimAI workspace.
            minimize: Required if no ``maximize`` parameter is defined. A list of one global coefficient to minimize.
                This global coefficient must correspond to one of the existing coefficients defined in your model configuration.
            maximize: Required if no ``minimize`` parameter is defined. A list of one global coefficient to maximize.
                This global coefficient must correspond to one of the existing coefficients defined in your model configuration.

                .. note::

                    - ``minimize`` and ``maximize`` are mutually exclusive objectives; define only one.
                    - The defined objective must be computed from the surface fields because mesh nodes must be involved.

            max_displacement: Optional. User-defined constraint on the maximum allowable deformation of the initial mesh in non-parametric optimization.
                It is specified as a list (``max_displacement``) matching the number of bounding boxes (``bounding_boxes``).

                For example, for two bounding boxes:

                - ``bounding_boxes = [[0,1,0,2,0,4],[10,2,10,4,10,5]]``
                - ``max_displacement = [0.002, 0.001]``

                Each value limits the displacement within the corresponding bounding box, using the same metric as the bounding box coordinates.
            show_progress: Optional. Whether to print progress bar on stdout.
                It is updated each time a new iteration is completed.
            boundary_conditions: Optional. **(Deprecated)** The values of the boundary conditions to perform the optimization at.
                The values must correspond to existing boundary conditions already defined in your SimAI workspace.

        Example:
          .. code-block:: python

            import ansys.simai.core as asc

            simai_client = asc.from_config(workspace="optimization-workspace")
            geometry = simai_client.geometries.list()[0]

            simai_client.optimizations.run_non_parametric(
                geometry,
                bounding_boxes=[[0, 1, 0, 1, 0, 1]],
                scalars={"VelocityX": 10.5},
                symmetries=["y"],
                n_iters=10,
                minimize=["TotalForceX"],
                show_progress=True,
            )

        Returns:
            An object containing the results of the optimization.

        """
        if scalars is None and boundary_conditions is not None:
            logger.warning(
                "The 'boundary_conditions' parameter is deprecated and will be removed in a future release. Please use the 'scalars' parameter instead."
            )
            scalars = boundary_conditions
        _validate_n_iters(n_iters)
        _validate_global_coefficients_for_non_parametric(minimize, maximize)
        _validate_bounding_boxes(bounding_boxes)
        _validate_max_displacement(max_displacement, bounding_boxes)
        _validate_axial_symmetry(axial_symmetry, symmetries)
        geometry = get_object_from_identifiable(geometry, self._client._geometry_directory)
        objective = _build_objective(minimize, maximize)
        optimization_parameters = {
            "boundary_conditions": scalars or {},
            "n_iters": n_iters,
            "objective": objective,
            "type": "non_parametric",
            "geometry_generation": {
                "geometry": geometry.id,
                "box_bounds_list": bounding_boxes,
                "symmetries": symmetries or [],
                "max_displacement": max_displacement,
                "axial_symmetry": axial_symmetry,
            },
        }
        with tqdm(total=n_iters, disable=not show_progress) as progress_bar:
            progress_bar.set_description("Creating optimization definition.")
            optimization = self._model_from(
                self._client._api.define_optimization(
                    geometry._fields["workspace_id"], optimization_parameters
                )
            )
            optimization.wait()
            trial_result = {"objectives": [], "geometries": []}
            logger.debug("Optimization defined. Starting optimization loop.")
            with keep.running(on_fail="warn"):
                for _ in range(n_iters):
                    logger.debug("Running iteration")
                    progress_bar.set_description("Running iteration")
                    trial_run = optimization._run_iteration({})
                    trial_run.wait()
                    progress_bar.set_postfix(
                        {
                            "objective": trial_run.fields["outcome_values"],
                        }
                    )
                    if trial_run.fields.get("is_feasible", True):
                        trial_result["objectives"].append(trial_run.fields["outcome_values"])
                        trial_result["geometries"].append(trial_run.fields["geometry"])
                    else:
                        logger.debug("Trial run results did not match constraints. Skipping.")
                    logger.debug("Trial complete.")
                    progress_bar.update(1)
                logger.debug("Optimization complete.")
                progress_bar.set_description("Optimization complete.")

            optimization_result = OptimizationResult(
                optimization, trial_result["objectives"], trial_result["geometries"]
            )
            return optimization_result


def _validate_bounding_boxes(bounding_boxes: list[list[float]]) -> None:
    if not isinstance(bounding_boxes, list) or not bounding_boxes:
        raise InvalidArguments("bounding_boxes must be a non-empty list.")

    for i, box in enumerate(bounding_boxes):
        if not isinstance(box, list) or len(box) != 6:
            raise InvalidArguments(
                f"Bounding box at index {i} should be a list of 6 values "
                f"[xmin, xmax, ymin, ymax, zmin, zmax]."
            )

        if not all(isinstance(value, (int, float)) for value in box):
            raise InvalidArguments(f"Bounding box at index {i} contains non-numeric values.")

        dimensions = ("x", "y", "z")
        for j in range(3):
            min_index = j * 2
            max_index = min_index + 1
            if not (box[min_index] < box[max_index]):
                raise InvalidArguments(
                    f"Bounding box at index {i}: {dimensions[j]}min ({box[min_index]}) "
                    f"must be less than {dimensions[j]}max ({box[max_index]})."
                )


def _validate_global_coefficients_for_non_parametric(minimize: List[str], maximize: List[str]):
    if minimize and maximize:
        raise InvalidArguments(
            "Only one of minimize or maximize can be provided for non parametric optimization."
        )

    active_list = minimize if minimize else maximize
    if not isinstance(active_list, list) or (
        isinstance(active_list, list) and len(active_list) != 1
    ):
        raise InvalidArguments(
            "minimize or maximize must be a list of one string for non parametric optimization."
        )


def _validate_outcome_constraints(outcome_constraints: list) -> None:
    if not isinstance(outcome_constraints, list):
        raise InvalidArguments("outcome_constraints must be a list")

    for i, constraint in enumerate(outcome_constraints):
        if not isinstance(constraint, str):
            raise InvalidArguments(f"Constraint at index {i} must be a string")

        # Check if constraint contains either >= or <=
        if ">=" in constraint:
            operator = ">="
        elif "<=" in constraint:
            operator = "<="
        else:
            raise InvalidArguments(
                f"Constraint at index {i} ({constraint}) must contain either >= or <= operator"
            )

        parts = constraint.split(operator)
        if len(parts) != 2:
            raise InvalidArguments(
                f"Constraint at index {i} ({constraint}) must be of form 'metric_name {operator} value'"
            )

        metric_name = parts[0].strip()
        value = parts[1].strip()

        if not metric_name:
            raise InvalidArguments(
                f"Constraint at index {i} ({constraint}) must have a metric name"
            )

        try:
            float(value)
        except ValueError as error:
            raise InvalidArguments(
                f"Constraint at index {i} ({constraint}): '{value}' must be a numeric value"
            ) from error


def _validate_n_iters(n_iters) -> None:
    if not isinstance(n_iters, int):
        raise InvalidArguments("n_iters must be an integer")

    if n_iters <= 0:
        raise InvalidArguments("n_iters must be strictly positive")


def _validate_max_displacement(
    max_displacement: Optional[List[float]], bounding_boxes: List[List[float]]
) -> None:
    if max_displacement is None:
        return

    if not isinstance(max_displacement, list):
        raise InvalidArguments("max_displacement must be a list")

    if not all(isinstance(value, (int, float)) for value in max_displacement):
        raise InvalidArguments("max_displacement contains non-numeric values")

    if len(max_displacement) != len(bounding_boxes):
        raise InvalidArguments(
            "Max displacement list and bounding boxes list must have the same number of items"
        )


def _validate_axial_symmetry(
    axial_symmetry: Optional[Literal["x", "y", "z"]],
    symmetries: Optional[List[Literal["x", "y", "z", "X", "Y", "Z"]]],
) -> None:
    if axial_symmetry is not None and axial_symmetry not in ("x", "y", "z"):
        raise InvalidArguments("axial_symmetry must be one of 'x', 'y', 'z' or None")

    if axial_symmetry is not None and symmetries is not None:
        raise InvalidArguments("symmetries and axial_symmetry are mutually exclusive parameters")


def _build_objective(minimize: list[str], maximize: list[str]) -> dict:
    if not minimize and not maximize:
        raise InvalidArguments("No global coefficient to optimize.")
    objective = {}
    if minimize:
        for global_coefficient in minimize:
            objective[global_coefficient] = {"minimize": True}
    if maximize:
        for global_coefficient in maximize:
            objective[global_coefficient] = {"minimize": False}
    return objective



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from ansys.simai.core.data.global_coefficients_requests import (
        ProcessGlobalCoefficient,
    )

METADATA_RAW = {
    "boundary_conditions": {
        "fields": [
            {
                "format": "value",
                "keys": None,
                "name": "Vx",
                "unit": None,
                "value": -5.569587230682373,
            }
        ]
    },
    "surface": {
        "fields": [
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "Pressure",
                "unit": None,
            },
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "TurbulentViscosity",
                "unit": None,
            },
        ]
    },
    "volume": {
        "fields": [
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "Pressure",
                "unit": None,
            }
        ]
    },
}


def test_process_formula_success(global_coefficient_request_factory, httpx_mock):
    gc_formula = "max(Pressure)"
    project_id = "xX007Xx"
    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/projects/{project_id}/process-formula",
        status_code=204,
    )
    process_gc: ProcessGlobalCoefficient = global_coefficient_request_factory(
        data={
            "id": f"{project_id}-process-{gc_formula}-cells",
        },
        project_id="xX007Xx",
        gc_formula=gc_formula,
        sample_metadata=METADATA_RAW,
        bc=["Vx"],
        surface_variables=["Pressure"],
        gc_location="cells",
    )
    check_sse_event = {
        "status": "successful",
        "target": {"action": "check", "formula": gc_formula, "gc_location": "cells"},
    }
    compute_result = 0.25478328
    compute_sse_event = {
        "status": "successful",
        "target": {"action": "compute", "formula": gc_formula, "gc_location": "cells"},
        "result": {"value": compute_result},
    }

    process_gc.run()

    assert not process_gc.is_ready

    process_gc._handle_job_sse_event(check_sse_event)

    assert not process_gc.is_ready
    assert process_gc.result is None

    process_gc._handle_job_sse_event(compute_sse_event)

    assert process_gc.is_ready
    assert process_gc.result == compute_result


def test_process_formula_success_with_cache(global_coefficient_request_factory, httpx_mock):
    gc_formula = "max(Pressure)"
    project_id = "xX007Xx"
    compute_result = 0.25478328
    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/projects/{project_id}/process-formula",
        status_code=200,
        json={"result": compute_result},
    )
    process_gc: ProcessGlobalCoefficient = global_coefficient_request_factory(
        data={
            "id": f"{project_id}-process-{gc_formula}",
        },
        project_id="xX007Xx",
        gc_formula=gc_formula,
        sample_metadata=METADATA_RAW,
        bc=["Vx"],
        surface_variables=["Pressure"],
        gc_location="cells",
    )

    process_gc.run()

    assert process_gc.is_ready
    assert process_gc.result == compute_result


@pytest.mark.parametrize("action", ["check", "compute"])
def test_process_formula_failure(global_coefficient_request_factory, action, httpx_mock):
    gc_formula = "max(Pressure)"
    project_id = "xX007Xx"
    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/projects/{project_id}/process-formula",
        status_code=204,
    )
    process_gc: ProcessGlobalCoefficient = global_coefficient_request_factory(
        request_type=action,
        data={
            "id": f"{project_id}-process-{gc_formula}",
        },
        project_id="xX007Xx",
        gc_formula=gc_formula,
        sample_metadata=METADATA_RAW,
        bc=["Vx"],
        surface_variables=["Pressure"],
        gc_location="cells",
    )
    reason_of_failure = "wrong shoes"
    sse_event = {
        "status": "failure",
        "target": {"action": action, "formula": gc_formula, "gc_location": "cells"},
        "reason": reason_of_failure,
    }

    process_gc.run()

    assert not process_gc.is_ready

    process_gc._handle_job_sse_event(sse_event)

    assert process_gc.has_failed
    assert all(keyword in process_gc.failure_reason for keyword in [reason_of_failure, gc_formula])



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import TYPE_CHECKING

from ansys.simai.core.data.model_configuration import ModelConfiguration

if TYPE_CHECKING:
    from ansys.simai.core.data.projects import Project
    from ansys.simai.core.data.workspaces import Workspace

METADATA_RAW = {
    "boundary_conditions": {
        "fields": [
            {
                "format": "value",
                "keys": None,
                "name": "Vx",
                "unit": None,
                "value": -5.569587230682373,
            }
        ]
    },
    "surface": {
        "fields": [
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "Pressure",
                "unit": None,
            },
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "TurbulentViscosity",
                "unit": None,
            },
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "WallShearStress_0",
                "unit": None,
            },
        ]
    },
    "volume": {
        "fields": [
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "Pressure",
                "unit": None,
            },
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "Velocity_0",
                "unit": None,
            },
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "Velocity_1",
                "unit": None,
            },
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "VolumeFractionWater",
                "unit": None,
            },
        ]
    },
}

MODEL_CONF_RAW = {
    "boundary_conditions": {"Vx": {}},
    "build_preset": "debug",
    "continuous": False,
    "fields": {
        "surface": [
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "Pressure",
                "unit": None,
            },
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "WallShearStress_0",
                "unit": None,
            },
        ],
        "surface_input": [],
        "surface_pp_input": [],
        "volume": [
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "Pressure",
                "unit": None,
            },
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "Velocity_0",
                "unit": None,
            },
        ],
    },
    "global_coefficients": [
        {"formula": "max(Pressure)", "name": "maxpress", "gc_location": "cells"}
    ],
    "scalars_to_predict": [],
    "simulation_volume": {
        "X": {"length": 300.0, "type": "relative_to_min", "value": 15.0},
        "Y": {"length": 80.0, "type": "absolute", "value": -80},
        "Z": {"length": 40.0, "type": "absolute", "value": -20.0},
    },
}

SAMPLE_RAW = {
    "extracted_metadata": METADATA_RAW,
    "id": "DarkKnight",
    "is_complete": True,
    "is_deletable": True,
    "is_in_a_project_being_trained": False,
    "is_sample_of_a_project": True,
    "luggage_version": "52.2.2",
}


def test_workspace_download_mer_data(simai_client, httpx_mock):
    """WHEN downloading mer csv file
    THEN the content of the file matches the content of the response.
    """

    workspace: Workspace = simai_client._workspace_directory._model_from(
        {"id": "0011", "name": "riri"}
    )
    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/workspaces/{workspace.id}/mer-data",
        content=b"mer-data-geometries",
        status_code=200,
    )

    in_memory = workspace.download_mer_data()
    data_in_file = in_memory.readline()
    assert data_in_file.decode("ascii") == "mer-data-geometries"


def test_workspace_rename(simai_client, httpx_mock):
    workspace: Workspace = simai_client._workspace_directory._model_from(
        {"id": "0011", "name": "riri"}
    )

    httpx_mock.add_response(
        method="PATCH",
        url="https://test.test/workspaces/0011",
        status_code=204,
    )
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/workspaces/0011",
        json={"id": "0011", "name": "fifi"},
        status_code=200,
    )

    workspace.rename("fifi")
    assert workspace.name == "fifi"


def test_get_workspace_model_configuration(mocker, simai_client, httpx_mock, training_data_factory):
    workspace: Workspace = simai_client._workspace_directory._model_from(
        {"id": "0011", "name": "riri", "project": "project101"}
    )

    raw_project = {
        "id": "project101",
        "name": "fifi",
        "sample": SAMPLE_RAW,
    }
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/projects/project101",
        json=raw_project,
        status_code=200,
    )
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/workspaces/0011/model/configuration",
        json=MODEL_CONF_RAW,
        status_code=200,
    )

    project: Project = simai_client._project_directory._model_from(raw_project)
    mocker.patch.object(project, "process_gc_formula", autospec=True)
    assert isinstance(workspace.model_configuration, ModelConfiguration)
    assert workspace.model_configuration.project == project
    assert workspace.model_configuration._to_payload() == MODEL_CONF_RAW


def test_workspace_list_predictions(simai_client, httpx_mock):
    workspace: Workspace = simai_client._workspace_directory._model_from(
        {"id": "0011", "name": "riri"}
    )

    raw_predictions = [
        {"id": "pred1"},
        {"id": "pred2"},
    ]
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/predictions/?workspace=0011",
        json=raw_predictions,
        status_code=200,
    )

    predictions = workspace.list_predictions()
    assert len(predictions) == 2
    assert predictions[0].id == "pred1"
    assert predictions[1].id == "pred2"


def test_workspace_list_geometries(simai_client, httpx_mock):
    workspace: Workspace = simai_client._workspace_directory._model_from(
        {"id": "0011", "name": "riri"}
    )

    raw_geometries = [
        {"id": "geom1"},
        {"id": "geom2"},
    ]
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/geometries/?workspace=0011",
        json=raw_geometries,
        status_code=200,
    )

    geometries = workspace.list_geometries()
    assert len(geometries) == 2
    assert geometries[0].id == "geom1"
    assert geometries[1].id == "geom2"



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import contextlib
import ssl
import subprocess
import sys
import threading
from http.server import HTTPServer, SimpleHTTPRequestHandler
from unittest.mock import patch

import pytest

import ansys.simai.core.errors as err
from ansys.simai.core import SimAIClient

from .conftest import disable_http_retry


#
# SETUP
#
@pytest.fixture(scope="module")
def tls_root_certificate(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("tls_root_certificate")
    # Create OpenSSL config file for the CA with key usage extension
    with open(tmp_path / "ca_openssl.cnf", "w") as f:
        f.write("""
[req]
distinguished_name = req_distinguished_name
req_extensions = v3_ca
x509_extensions = v3_ca

[req_distinguished_name]

[v3_ca]
subjectKeyIdentifier = hash
authorityKeyIdentifier = keyid:always,issuer:always
basicConstraints = critical, CA:true
keyUsage = critical, digitalSignature, cRLSign, keyCertSign
""")
    # Create CA key and certificate with extensions
    subprocess.run(["openssl", "req", "-x509", "-newkey", "rsa:2048", "-nodes", "-keyout", tmp_path / "ca_key.pem", "-out", tmp_path / "test_ca.pem", "-days", "3650", "-subj", "/CN=Test CA", "-config", tmp_path / "ca_openssl.cnf"], check=False)  # fmt: skip # noqa: S607, S603
    # Create server key and certificate
    subprocess.run(["openssl", "req", "-newkey", "rsa:2048", "-nodes", "-keyout", tmp_path / "server_key.pem", "-out", tmp_path / "server_csr.pem", "-subj", "/CN=localhost"], check=False)  # fmt: skip # noqa: S607, S603
    with open(tmp_path / "server_openssl.cnf", "w") as f:
        # Create a config file for OpenSSL to include the SAN and key usage extensions
        f.write("""
[req]
distinguished_name = req_distinguished_name

[req_distinguished_name]

[server_ext]
basicConstraints = CA:FALSE
keyUsage = critical, digitalSignature, keyEncipherment
extendedKeyUsage = serverAuth
subjectAltName = DNS:localhost
""")
    subprocess.run(["openssl", "x509", "-req", "-in", tmp_path / "server_csr.pem", "-CA", tmp_path / "test_ca.pem", "-CAkey", tmp_path / "ca_key.pem", "-CAcreateserial", "-out", tmp_path / "server_cert.pem", "-days", "3650", "-extensions", "server_ext", "-extfile", tmp_path / "server_openssl.cnf"], check=False)  # fmt: skip # noqa: S607, S603

    return {
        "ca": tmp_path / "test_ca.pem",
        "server_key": tmp_path / "server_key.pem",
        "server_cert": tmp_path / "server_cert.pem",
    }


@pytest.fixture(scope="module")
def https_server(tls_root_certificate):
    class SimpleHTTPSHandler(SimpleHTTPRequestHandler):
        def do_GET(self):
            self.send_response(200)
            self.send_header("Content-type", "text/plain")
            self.end_headers()
            self.wfile.write(b"Hello, Secure World!")

    ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ssl_context.load_cert_chain(
        certfile=tls_root_certificate["server_cert"], keyfile=tls_root_certificate["server_key"]
    )
    httpd = HTTPServer(("localhost", 48219), SimpleHTTPSHandler)
    httpd.socket = ssl_context.wrap_socket(
        httpd.socket,
        server_side=True,
    )
    server_thread = threading.Thread(target=httpd.serve_forever, daemon=True)
    server_thread.start()
    yield "https://localhost:48219"
    httpd.shutdown()


BASE_CLT_ARGS = {
    "url": "https://test.example.com",
    "organization": "dummy",
    "_disable_authentication": True,
    "no_sse_connection": True,
    "skip_version_check": True,
}


#
# TESTS
#
@pytest.mark.skipif(
    sys.version_info < (3, 11),
    reason='"system" requires Python >= 3.10, "patch" is broken in python 3.10',
)
def test_client_config_tls_ca_bundle_system(tls_root_certificate, https_server):
    clt = SimAIClient(**BASE_CLT_ARGS, tls_ca_bundle="system")
    with disable_http_retry(clt._api, https_server):
        # The system CA rejects the test CA by default
        with pytest.raises(err.ConnectionError):
            clt._api._get(https_server)

        # If the host system trusts the test CA, pysimai trusts it !
        @contextlib.contextmanager
        def load_test_ca_as_truststore_system_ca(ctx: "ssl.SSLContext"):
            ctx.load_verify_locations(cafile=tls_root_certificate["ca"])
            yield

        with patch(
            "truststore._api._configure_context", side_effect=load_test_ca_as_truststore_system_ca
        ):
            clt._api._get(https_server, return_json=False)


@pytest.mark.skipif(
    sys.version_info >= (3, 10), reason="The error is only raised for python < 3.10"
)
def test_client_config_tls_ca_bundle_system_on_usupported_python_version():
    with pytest.raises(err.ConfigurationError, match="python >= 3.10"):
        SimAIClient(**BASE_CLT_ARGS, tls_ca_bundle="system")


def test_client_config_tls_ca_bundle_unsecure_none(https_server):
    clt = SimAIClient(**BASE_CLT_ARGS, tls_ca_bundle="unsecure-none")
    clt._api._get(https_server, return_json=False)


@pytest.mark.skipif(
    sys.platform == "win32", reason="tls_root_certificate fixture doesn't work on windows ?"
)
def test_client_config_tls_ca_bundle_path(tls_root_certificate, https_server):
    clt = SimAIClient(**BASE_CLT_ARGS, tls_ca_bundle=tls_root_certificate["ca"])
    clt._api._get(https_server, return_json=False)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from functools import partial
from typing import TYPE_CHECKING

import pytest

from ansys.simai.core.data.models import ModelConfiguration
from ansys.simai.core.data.training_data import TrainingData
from ansys.simai.core.errors import ApiClientError, ProcessingError

if TYPE_CHECKING:
    from ansys.simai.core.data.projects import Project


def test_project_rename(simai_client, httpx_mock):
    project: Project = simai_client._project_directory._model_from({"id": "0011", "name": "riri"})

    httpx_mock.add_response(
        method="PATCH",
        url="https://test.test/projects/0011",
        status_code=204,
    )
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/projects/0011",
        json={"id": "0011", "name": "fifi"},
        status_code=200,
    )

    project.rename("fifi")
    assert project.name == "fifi"


def test_project_list_training_data(simai_client, httpx_mock):
    project: Project = simai_client._project_directory._model_from({"id": "0011", "name": "riri"})

    httpx_mock.add_response(
        is_reusable=True,
        method="GET",
        url="https://test.test/projects/0011/data",
        headers={"Link": '<https://test.test/projects/0011/data?last_id=first>; rel="next"'},
        json=[{"id": "first"}],
        status_code=200,
    )

    httpx_mock.add_response(
        is_reusable=True,
        method="GET",
        url="https://test.test/projects/0011/data?last_id=first",
        json=[{"id": "second"}],
        status_code=200,
    )

    tds = project.list_training_data()

    # deprecated
    assert [data.id for data in project.data] == ["first", "second"]

    assert [data.id for data in tds] == ["first", "second"]


def test_project_sample(simai_client, httpx_mock):
    raw_td = {"id": "28-06-1712", "name": "jean-jacques rousseau"}
    raw_project = {"id": "xX007Xx", "name": "fifi", "sample": None}

    project: Project = simai_client._project_directory._model_from(raw_project)

    httpx_mock.add_response(
        method="PUT",
        url=f"https://test.test/projects/{project.id}/sample",
        status_code=200,
    )
    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/projects/{project.id}",
        json={**raw_project, "sample": raw_td},
        status_code=200,
    )
    assert project.sample is None
    project.sample = raw_td["id"]
    assert isinstance(project.sample, TrainingData)
    assert project.sample.id == raw_td["id"]


@pytest.mark.parametrize(
    "status_code,response_body,error_type",
    [
        (200, {"is_trainable": True}, None),
        (200, {"is_trainable": False, "reason": "a reason why it fails"}, None),
        (400, None, ApiClientError),
    ],
)
def test_is_trainable(simai_client, status_code, response_body, error_type, httpx_mock):
    """Test is_trainable method according to is_trainable parameter and the response status code."""
    raw_project = {"id": "xX007Xx", "name": "fifi", "sample": None}

    project: Project = simai_client._project_directory._model_from(raw_project)

    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/projects/{project.id}/trainable",
        status_code=status_code,
        json=response_body,
    )

    if status_code == 200:
        pt = project.is_trainable()

        assert bool(pt) is response_body.get("is_trainable")
        assert pt.reason == response_body.get("reason")
    else:
        with pytest.raises(error_type):
            project.is_trainable()


########## Variables for testing get_variables ##########
METADATA_RAW = {
    "boundary_conditions": {
        "fields": [
            {
                "format": "value",
                "keys": None,
                "name": "Vx",
                "unit": None,
                "value": -5.569587230682373,
            }
        ]
    },
    "surface": {
        "fields": [
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "Pressure",
                "unit": None,
            },
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "TurbulentViscosity",
                "unit": None,
            },
        ]
    },
    "volume": {
        "fields": [
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "Pressure",
                "unit": None,
            }
        ]
    },
}

VARS_OUT = {
    "boundary_conditions": ["Vx"],
    "surface": ["Pressure", "TurbulentViscosity"],
    "volume": ["Pressure"],
}

SAMPLE_RAW = {
    "extracted_metadata": METADATA_RAW,
    "id": "DarkKnight",
    "is_complete": True,
    "is_deletable": True,
    "is_in_a_project_being_trained": False,
    "is_sample_of_a_project": True,
    "luggage_version": "52.2.2",
}


@pytest.mark.parametrize(
    "in_sample,vars_out",
    [
        (None, None),
        (SAMPLE_RAW, VARS_OUT),
    ],
)
def test_get_variables(simai_client, in_sample, vars_out):
    """WHEN sample exists
    THEN the output is the names of the available variables"""
    raw_project = {"id": "xX007Xx", "name": "fifi", "sample": in_sample}

    project: Project = simai_client._project_directory._model_from(raw_project)

    var_pool = project.get_variables()
    assert var_pool == vars_out


def test_last_model_configuration(simai_client):
    """Test last_configuration property."""

    last_conf = {
        "boundary_conditions": {"Vx": {}},
        "build_preset": "debug",
        "continuous": False,
        "fields": {
            "surface": [
                {
                    "format": "value",
                    "keys": None,
                    "location": "cell",
                    "name": "TurbulentViscosity",
                    "unit": None,
                }
            ],
            "surface_input": [],
            "surface_pp_input": [],
            "volume": [
                {
                    "format": "value",
                    "keys": None,
                    "location": "cell",
                    "name": "Pressure",
                    "unit": None,
                }
            ],
        },
        "global_coefficients": [],
        "simulation_volume": {
            "X": {"length": 300.0, "type": "relative_to_min", "value": 15.0},
            "Y": {"length": 80.0, "type": "absolute", "value": -80},
            "Z": {"length": 40.0, "type": "absolute", "value": -20.0},
        },
        "scalars_to_predict": [],
    }

    raw_project = {
        "id": "xX007Xx",
        "name": "fifi",
        "sample": SAMPLE_RAW,
        "last_model_configuration": last_conf,
    }

    project: Project = simai_client._project_directory._model_from(raw_project)

    project_last_conf = project.last_model_configuration

    assert isinstance(project_last_conf, ModelConfiguration)
    assert project_last_conf._to_payload() == last_conf


def test_process_gc_formula_with_no_sample(simai_client):
    """WHEN no sample is defined in the project
    THEN an exception is raised in process_gc_formula."""
    raw_project = {"id": "xX007Xx", "name": "fifi", "sample": None}
    project: Project = simai_client._project_directory._model_from(raw_project)

    with pytest.raises(ProcessingError):
        project.process_gc_formula("max(Pressure)")


def test_process_gc_formula_without_cache(simai_client, delayed_events, httpx_mock):
    gc_formula = "max(Pressure)"
    compute_result = 0.25478328
    raw_project = {"id": "xX007Xx", "name": "fifi", "sample": SAMPLE_RAW}
    project: Project = simai_client._project_directory._model_from(raw_project)
    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/projects/{project.id}/process-formula",
        status_code=204,
    )
    check_sse_event = {
        "status": "successful",
        "target": {
            "id": "xX007Xx",
            "action": "check",
            "formula": gc_formula,
            "gc_location": "cells",
        },
    }
    compute_sse_event = {
        "status": "successful",
        "target": {
            "id": "xX007Xx",
            "action": "compute",
            "formula": gc_formula,
            "location": "cells",
            "gc_location": "cells",
        },
        "result": {"value": compute_result},
    }
    sse_callback = simai_client._process_gc_formula_directory._handle_sse_event
    # Schedule events with proper sequencing - check event first, then compute event
    delayed_events.add(partial(sse_callback, data=check_sse_event))
    delayed_events.add(partial(sse_callback, data=compute_sse_event))
    delayed_events.start()

    result = project.process_gc_formula(gc_formula)

    assert result == compute_result


def test_process_gc_formula_with_cache(simai_client, httpx_mock):
    raw_project = {"id": "xX007Xx", "name": "fifi", "sample": SAMPLE_RAW}
    project: Project = simai_client._project_directory._model_from(raw_project)
    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/projects/{project.id}/process-formula",
        status_code=200,
        json={
            "formula": "max(Pressure)",
            "result": 0.25478328,
        },
    )

    # No need to wait for the SSE event, the cache value is returned by the API
    result = project.process_gc_formula("max(Pressure)")

    assert result == 0.25478328


def test_cancel_build_for_specified_project_id(simai_client, httpx_mock):
    """WHEN I call projects.cancel_build() for a specified project id
    THEN it returns None if there is an build in progress
    AND raises a ProcessingError otherwise
    """

    project = simai_client.projects._model_from({"id": "e45y123", "name": "proj"})

    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/projects/{project.id}",
        json={"id": "e45y123", "is_being_trained": True},
        status_code=200,
    )

    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/projects/{project.id}/cancel-training",
        json={"id": "e45y123", "is_being_trained": False},
        status_code=200,
    )

    simai_client.projects.cancel_build(project.id)

    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/projects/{project.id}",
        json={"id": "e45y123", "is_being_trained": False},
        status_code=200,
    )

    with pytest.raises(ProcessingError) as excinfo:
        simai_client.projects.cancel_build(project.id)
        assert "No build pending for this project" in excinfo.value


def test_cancel_active_build_from_project(simai_client, httpx_mock):
    """WHEN I call cancel_build() from a project with an active build
    THEN returns None
    """

    project = simai_client.projects._model_from({"id": "e45y123", "name": "proj"})

    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/projects/{project.id}",
        json={"id": "e45y123", "is_being_trained": True},
        status_code=200,
    )

    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/projects/{project.id}/cancel-training",
        json={"id": "e45y123"},
        status_code=200,
    )

    project.cancel_build()
    assert len(httpx_mock.get_requests()) == 2


def test_cancel_inactive_build_from_project(simai_client, httpx_mock):
    """WHEN I call cancel_build() from a project with an inactive build
    THEN it raises a ProcessingError
    """

    project = simai_client.projects._model_from({"id": "e45y123", "name": "proj"})

    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/projects/{project.id}",
        json={"id": "e45y123", "is_being_trained": False},
        status_code=200,
    )

    with pytest.raises(ProcessingError) as excinfo:
        project.cancel_build()
        assert "No build pending for this project" in excinfo.value
    assert len(httpx_mock.get_requests()) == 1


def test_project_list_workspaces(simai_client, httpx_mock):
    project = simai_client.projects._model_from({"id": "0011", "name": "riri"})

    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/projects/{project.id}/workspaces",
        json=[{"id": "ws01", "name": "Workspace01"}, {"id": "ws02", "name": "Workspace02"}],
        status_code=200,
    )

    workspaces = project.list_workspaces()

    assert len(workspaces) == 2
    assert workspaces[0].id == "ws01"
    assert workspaces[1].id == "ws02"


def test_project_list_models(simai_client, httpx_mock):
    project = simai_client.projects._model_from({"id": "0011", "name": "riri"})

    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/projects/{project.id}/models",
        json=[{"id": "model01", "name": "Model01"}, {"id": "model02", "name": "Model02"}],
        status_code=200,
    )

    models = project.list_models()

    assert len(models) == 2
    assert models[0].id == "model01"
    assert models[1].id == "model02"



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json

import httpx
import pytest

from ansys.simai.core.data.selections import Selection
from ansys.simai.core.errors import MultipleErrors


@pytest.fixture()
def four_geometries_test_set(geometry_factory, prediction_factory):
    """A list of 4 geometries with xbow in [21, 23, 25, 27]"""
    geometries = []
    for i in [21, 23, 25, 27]:
        if i == 21:
            predictions = [
                prediction_factory(id="pred19.8", boundary_conditions={"Vx": 19.8}),
                prediction_factory(id="pred20.2", boundary_conditions={"Vx": 20.2}),
                prediction_factory(id="pred20.5", boundary_conditions={"Vx": 20.5}),
            ]
        else:
            predictions = []
        geometry = geometry_factory(id=f"xbow-{i}", predictions=predictions, metadata={"xbow": i})
        geometries.append(geometry)
    yield geometries


def test_selection_content(four_geometries_test_set):
    """WHEN creating a Selection from a list of geometries and scalars
    THEN the selection contains all geometry—boundary-condition combinations
    """
    speeds = [20.2, 20.4, 20.6]
    scalars = [{"Vx": v} for v in speeds]
    selection = Selection(four_geometries_test_set, scalars)
    # 4 geometries * 3 scalars
    points = selection.points
    assert len(points) == 12

    received_xbow_speed_combinations = {
        (p.geometry.metadata["xbow"], p.scalars["Vx"]) for p in points
    }
    expected_combinations = set()
    for speed in speeds:
        for xbow in [21, 23, 25, 27]:
            expected_combinations.add((xbow, speed))
    assert received_xbow_speed_combinations == expected_combinations


def test_selection_get_predictions(four_geometries_test_set):
    """WHEN accessing the selection's predictions
    THEN predictions existing for the geometry­—boundary-condition couple are returned
    """
    scalars = [{"Vx": v} for v in [20.2, 20.4, 20.6]]
    selection = Selection(four_geometries_test_set, scalars)

    predictions = selection.get_predictions()
    assert len(predictions) == 1
    assert predictions[0].id == "pred20.2"

    points_with_prediction = [p for p in selection.points if p.prediction is not None]
    assert len(points_with_prediction) == 1
    point = points_with_prediction[0]
    assert point.geometry.id == "xbow-21"
    assert point.scalars == {"Vx": 20.2}


def test_selection_run_predictions(geometry_factory, prediction_factory, httpx_mock):
    """WHEN calling run_prediction() on a selection
    THEN a POST request is launched for each not-existing prediction
    AND after the call, selection.predictions contains values for
        all combinations of geometries and scalar
    """
    # We create 2 geometries, geom 77777 having 1 prediction, geom 88888 having none
    geometries = [
        geometry_factory(
            id="77777",
            predictions=[prediction_factory(id="k6k6k6", boundary_conditions={"Vx": 4.5})],
        ),
        geometry_factory(id="88888", predictions=[]),
    ]

    # Select 3 speeds
    interesting_speeds_x = {4.5, 5.5, 6.5}
    scalars = [{"Vx": v} for v in interesting_speeds_x]

    selection = Selection(geometries, scalars)

    # Selection contains 6 points (2 geom * 2 BC)
    assert len(selection.points) == 6
    # Selection contains only 1 prediction (geom 0001 - speed 4.5)
    assert len(selection.predictions) == 1
    assert selection.predictions[0].scalars["Vx"] == 4.5
    assert len(selection.get_runnable_predictions()) == 5

    # Endpoints for prediction creations
    def geometry1_pred_request_callback(request):
        payload = json.loads(request.content)
        scalars = payload["boundary_conditions"]
        # geometry1 had a pred for speed 4.5, it should not be recreated
        assert scalars["Vx"] in {5.5, 6.5}
        return httpx.Response(
            200,
            json=(
                {
                    "id": f"pred{scalars}",
                    "status": "queued",
                    "boundary_conditions": scalars,
                }
            ),
        )

    def geometry2_pred_request_callback(request):
        payload = json.loads(request.content)
        scalars = payload["boundary_conditions"]
        # geometry2 had no pred, preds for all speed will be created
        assert scalars["Vx"] in {4.5, 5.5, 6.5}
        return httpx.Response(
            200,
            json=(
                {
                    "id": f"pred{scalars}",
                    "status": "queued",
                    "boundary_conditions": scalars,
                }
            ),
        )

    for _ in range(2):
        httpx_mock.add_callback(
            geometry1_pred_request_callback,
            method="POST",
            url="https://test.test/geometries/77777/predictions",
        )
    for _ in range(3):
        httpx_mock.add_callback(
            geometry2_pred_request_callback,
            method="POST",
            url="https://test.test/geometries/88888/predictions",
        )

    # Let's run all predictions in the selection
    selection.run_predictions()

    # Now selection contains 6 predictions, for both geometries and all expected speeds
    assert len(selection.points) == 6
    assert len(selection.predictions) == 6

    expected_combinations = set()
    for geom in geometries:
        for speed_x in interesting_speeds_x:
            expected_combinations.add((geom.id, speed_x))
    received_combinations = {
        (p.geometry.id, p.scalars["Vx"]) for p in selection.points_with_prediction if p is not None
    }
    assert expected_combinations == received_combinations

    assert len(selection.get_runnable_predictions()) == 0


def test_selection_tolerance(geometry_factory, prediction_factory):
    """WHEN creating a selection with the default tolerance
    THEN an prediction with a scalar differing from less than
        10**-6 is included in the selection
    """
    small_epsilon = 10**-8
    big_epsilon = 10**-5
    geometry = geometry_factory(
        predictions=[
            prediction_factory(id="here-i-am", boundary_conditions={"Vx": 11 + small_epsilon}),
            prediction_factory(id="i-am-not", boundary_conditions={"Vx": 11 + big_epsilon}),
        ],
    )
    scalars = [{"Vx": 11}, {"Vx": 12}]

    selection = Selection([geometry], scalars)
    predictions = selection.get_predictions()
    assert len(predictions) == 1
    assert predictions[0].id == "here-i-am"


def test_selection_run_prediction_error(geometry_factory, httpx_mock):
    """WHEN calling run_predictions, and some calls return an error
    THEN all the predictions are ran nevertheless
    AND at the end a MultipleErrors is raised
    """
    geometry = geometry_factory()
    speeds = [-11, -10, -9]
    selection = Selection([geometry], [{"Vx": v} for v in speeds])

    nb_calls = 0

    def pred_creation_callback(request):
        nonlocal nb_calls
        nb_calls += 1
        payload = json.loads(request.content)
        scalars = payload["boundary_conditions"]
        # send 422 error for the first 2 calls, 200 for the last one
        if nb_calls < 3:
            return httpx.Response(
                422,
                json=(
                    {
                        "code": 422,
                        "errors": {"json": {"speed": "Planets are not aligned"}},
                        "status": "Stellar Entity",
                    }
                ),
            )
        else:
            return httpx.Response(
                200,
                json=(
                    {
                        "id": "saturn",
                        "status": "queued",
                        "boundary_conditions": scalars,
                    }
                ),
            )

    for _ in range(3):
        httpx_mock.add_callback(
            pred_creation_callback,
            method="POST",
            url=f"https://test.test/geometries/{geometry.id}/predictions",
        )

    with pytest.raises(MultipleErrors):
        selection.run_predictions()

    # assert 3 calls have been made despite the first two having failed
    assert len(httpx_mock.get_requests()) == 3
    assert len(selection.predictions) == 1
    assert selection.predictions[0].id == "saturn"


def test_selection_parameters(four_geometries_test_set):
    """WHEN creating a Selection
    IF I don't pass a Geometry or list thereof, and a Scalar or list thereof
    THEN a TypeError is raised
    """

    scalars = [{"Vx": v} for v in [-11.9, -5, 0, 50, 900.4]]

    Selection(four_geometries_test_set, scalars)
    Selection(four_geometries_test_set[0], scalars)
    Selection(four_geometries_test_set, scalars[0])
    Selection(four_geometries_test_set[0], scalars[0])
    Selection(four_geometries_test_set, {"Vx": 1})

    with pytest.raises(TypeError):
        Selection(scalars, four_geometries_test_set)
    with pytest.raises(TypeError):
        Selection(None, scalars)
    with pytest.raises(TypeError):
        Selection(12.5, scalars)

    with pytest.raises(TypeError):
        Selection(four_geometries_test_set, [{"Vx": "a", "Vy": "b", "Vz": "c"}])
    with pytest.raises(TypeError):
        Selection(four_geometries_test_set, [(1, 0, 0, 0)])

    with pytest.raises(TypeError):
        Selection(four_geometries_test_set, 1)
    with pytest.raises(TypeError):
        Selection(four_geometries_test_set, 1, 0, 0)
    with pytest.raises(TypeError):
        Selection(four_geometries_test_set, scalars, tolerance="")
    with pytest.raises(ValueError):
        Selection(four_geometries_test_set, scalars, tolerance=-10)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from datetime import datetime, timezone

import pytest

from ansys.simai.core.data.current_user import Consent
from ansys.simai.core.errors import NotFoundError


def test_list_consents(simai_client, httpx_mock):
    """WHEN listing consents
    THEN the API is called and consents are returned as Consent objects.
    """
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/users/offline-tokens",
        json=[
            {
                "client_id": "sdk",
                "created_date": "2025-01-15T10:30:00+00:00",
                "last_updated_date": "2025-01-16T14:20:00+00:00",
                "granted_scopes": ["openid", "offline_access"],
            },
            {
                "client_id": "webapp",
                "created_date": "2025-01-10T08:00:00+00:00",
                "last_updated_date": "2025-01-10T08:00:00+00:00",
                "granted_scopes": ["openid"],
            },
        ],
        status_code=200,
    )

    consents = simai_client.me.consents.list()

    assert len(consents) == 2
    assert all(isinstance(c, Consent) for c in consents)

    assert consents[0].client_id == "sdk"
    assert consents[0].created_date == datetime(2025, 1, 15, 10, 30, 0, tzinfo=timezone.utc)
    assert consents[0].last_updated_date == datetime(2025, 1, 16, 14, 20, 0, tzinfo=timezone.utc)
    assert consents[0].granted_scopes == ["openid", "offline_access"]

    assert consents[1].client_id == "webapp"
    assert consents[1].granted_scopes == ["openid"]


def test_list_consents_empty(simai_client, httpx_mock):
    """WHEN listing consents and user has none
    THEN an empty list is returned.
    """
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/users/offline-tokens",
        json=[],
        status_code=200,
    )

    consents = simai_client.me.consents.list()

    assert consents == []


def test_revoke_consent(simai_client, httpx_mock):
    """WHEN revoking a consent
    THEN the API is called with the correct client_id.
    """
    httpx_mock.add_response(
        method="DELETE",
        url="https://test.test/users/offline-tokens/sdk",
        status_code=204,
    )

    simai_client.me.consents.revoke("sdk")

    request = httpx_mock.get_request()
    assert request.method == "DELETE"
    assert str(request.url) == "https://test.test/users/offline-tokens/sdk"


def test_revoke_consent_not_found(simai_client, httpx_mock):
    """WHEN revoking a non-existent consent
    THEN a NotFoundError is raised.
    """
    httpx_mock.add_response(
        method="DELETE",
        url="https://test.test/users/offline-tokens/nonexistent",
        json={"message": "No consent found for client: nonexistent"},
        status_code=404,
    )

    with pytest.raises(NotFoundError):
        simai_client.me.consents.revoke("nonexistent")



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import TYPE_CHECKING

import pytest

from ansys.simai.core.data.geomai.models import GeomAIModelConfiguration
from ansys.simai.core.errors import ProcessingError

if TYPE_CHECKING:
    from ansys.simai.core.data.geomai.projects import GeomAIProject


def test_project_rename(simai_client, httpx_mock):
    project: GeomAIProject = simai_client.geomai._project_directory._model_from(
        {"id": "0011", "name": "riri"}
    )

    httpx_mock.add_response(
        method="PATCH",
        url="https://test.test/geomai/projects/0011",
        status_code=204,
    )
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/geomai/projects/0011",
        json={"id": "0011", "name": "fifi"},
        status_code=200,
    )

    project.rename("fifi")
    assert project.name == "fifi"


def test_project_list_training_data(simai_client, httpx_mock):
    project: GeomAIProject = simai_client.geomai._project_directory._model_from(
        {"id": "0011", "name": "riri"}
    )

    httpx_mock.add_response(
        is_reusable=True,
        method="GET",
        url="https://test.test/geomai/projects/0011/training-data",
        headers={
            "Link": '<https://test.test/geomai/projects/0011/training-data?last_id=first>; rel="next"'
        },
        json=[{"id": "first"}],
        status_code=200,
    )

    httpx_mock.add_response(
        is_reusable=True,
        method="GET",
        url="https://test.test/geomai/projects/0011/training-data?last_id=first",
        json=[{"id": "second"}],
        status_code=200,
    )
    workspaces = project.list_training_data()

    # deprecated
    assert [data.id for data in project.data()] == ["first", "second"]

    assert [data.id for data in workspaces] == ["first", "second"]


def test_last_model_configuration(simai_client):
    """Test last_configuration property."""

    last_conf = {"nb_latent_param": 3, "build_preset": "short", "nb_epochs": None}

    raw_project = {
        "id": "xX007Xx",
        "name": "fifi",
        "last_model_configuration": last_conf,
    }

    project: GeomAIProject = simai_client.geomai._project_directory._model_from(raw_project)

    project_last_conf = project.last_model_configuration

    assert isinstance(project_last_conf, GeomAIModelConfiguration)
    assert project_last_conf.model_dump() == last_conf


def test_cancel_build_for_specified_project_id(simai_client, httpx_mock):
    """WHEN I call projects.cancel_build() for a specified project id
    THEN it returns None if there is an build in progress
    AND raises a ProcessingError otherwise
    """

    project = simai_client.projects._model_from({"id": "e45y123", "name": "proj"})

    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/projects/{project.id}",
        json={"id": "e45y123", "is_being_trained": True},
        status_code=200,
    )

    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/projects/{project.id}/cancel-training",
        json={"id": "e45y123", "is_being_trained": False},
        status_code=200,
    )

    simai_client.projects.cancel_build(project.id)

    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/projects/{project.id}",
        json={"id": "e45y123", "is_being_trained": False},
        status_code=200,
    )

    with pytest.raises(ProcessingError) as excinfo:
        simai_client.projects.cancel_build(project.id)
        assert "No build pending for this project" in excinfo.value


def test_cancel_active_build_from_project(simai_client, httpx_mock):
    """WHEN I call cancel_build() from a project with an active build
    THEN returns None
    """

    project = simai_client.projects._model_from({"id": "e45y123", "name": "proj"})

    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/projects/{project.id}",
        json={"id": "e45y123", "is_being_trained": True},
        status_code=200,
    )

    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/projects/{project.id}/cancel-training",
        json={"id": "e45y123"},
        status_code=200,
    )

    project.cancel_build()
    assert len(httpx_mock.get_requests()) == 2


def test_cancel_inactive_build_from_project(simai_client, httpx_mock):
    """WHEN I call cancel_build() from a project with an inactive build
    THEN it raises a ProcessingError
    """

    project = simai_client.projects._model_from({"id": "e45y123", "name": "proj"})

    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/projects/{project.id}",
        json={"id": "e45y123", "is_being_trained": False},
        status_code=200,
    )

    with pytest.raises(ProcessingError) as excinfo:
        project.cancel_build()
        assert "No build pending for this project" in excinfo.value
    assert len(httpx_mock.get_requests()) == 1


def test_geomai_project_list_workspaces(simai_client, httpx_mock):
    project = simai_client.geomai.projects._model_from({"id": "0011", "name": "riri"})

    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/geomai/projects/{project.id}/workspaces",
        json=[{"id": "ws01", "name": "Workspace01"}, {"id": "ws02", "name": "Workspace02"}],
        status_code=200,
    )

    workspaces = project.list_workspaces()

    assert len(workspaces) == 2
    assert workspaces[0].id == "ws01"
    assert workspaces[1].id == "ws02"


def test_geomai_project_list_models(simai_client, httpx_mock):
    project = simai_client.geomai.projects._model_from({"id": "0011", "name": "riri"})

    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/geomai/projects/{project.id}/models",
        json=[{"id": "model01"}, {"id": "model02"}],
        status_code=200,
    )

    models = project.list_models()

    assert len(models) == 2
    assert models[0].id == "model01"
    assert models[1].id == "model02"



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ruff: noqa: E731

import json
import threading

import pytest
from httpx_sse import ServerSentEvent

from ansys.simai.core.data.optimizations import (
    _validate_axial_symmetry,
    _validate_bounding_boxes,
    _validate_global_coefficients_for_non_parametric,
    _validate_max_displacement,
    _validate_n_iters,
    _validate_outcome_constraints,
)
from ansys.simai.core.errors import InvalidArguments


def test_validate_bounding_boxes_success():
    bounding_boxes = [[0.1, 0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1.0, 1.1, 1.2]]

    _validate_bounding_boxes(bounding_boxes)


@pytest.mark.parametrize(
    "bounding_boxes, error_message",
    [
        ([], "bounding_boxes must be a non-empty list"),
        ([[0.1, 0.1, 0.1, 0.1]], "Bounding box at index 0 should be a list of 6 values"),
        (
            [[0.1, "foo", 0.1, 0.2, 0.1, 0.2]],
            "Bounding box at index 0 contains non-numeric values.",
        ),
        (
            [[0.1, 0.1, 0.1, 0.2, 0.1, 0.2]],
            r"Bounding box at index 0: xmin \(0\.1\) must be less than xmax \(0\.1\)",
        ),
        (
            [[0.1, 0.2, 0.1, 0.1, 0.1, 0.2]],
            r"Bounding box at index 0: ymin \(0\.1\) must be less than ymax \(0\.1\)",
        ),
        (
            [[0.1, 0.2, 0.1, 0.2, 0.1, 0.1]],
            r"Bounding box at index 0: zmin \(0\.1\) must be less than zmax \(0\.1\)",
        ),
    ],
)
def test_validate_bounding_boxes_fails(bounding_boxes, error_message):
    with pytest.raises(
        expected_exception=InvalidArguments,
        match=error_message,
    ):
        _validate_bounding_boxes(bounding_boxes)


def test_validate_global_coefficients_for_non_parametric_success():
    # Test with minimize only
    _validate_global_coefficients_for_non_parametric(minimize=["TotalForceX"], maximize=[])

    # Test with maximize only
    _validate_global_coefficients_for_non_parametric(minimize=[], maximize=["TotalForceY"])


@pytest.mark.parametrize(
    "minimize, maximize, error_message",
    [
        (
            ["TotalForceX"],
            ["TotalForceY"],
            "Only one of minimize or maximize can be provided for non parametric optimization.",
        ),
        (
            [],
            [],
            "minimize or maximize must be a list of one string for non parametric optimization.",
        ),
        (
            ["TotalForceX", "Pressure"],
            [],
            "minimize or maximize must be a list of one string for non parametric optimization.",
        ),
        (
            None,
            None,
            "minimize or maximize must be a list.",
        ),
        (
            {},
            [],
            "minimize or maximize must be a list of one string for non parametric optimization.",
        ),
        (
            [],
            None,
            "minimize or maximize must be a list of one string for non parametric optimization.",
        ),
        (
            "string_instead_of_list",
            [],
            "minimize or maximize must be a list of one string for non parametric optimization.",
        ),
    ],
)
def test_validate_global_coefficients_for_non_parametric_fails(minimize, maximize, error_message):
    with pytest.raises(
        expected_exception=InvalidArguments,
        match=error_message,
    ):
        _validate_global_coefficients_for_non_parametric(minimize, maximize)


def test_validate_outcome_constraints_success():
    outcome_constraints = ["TotalForceX <= 10", "Pressure >= 5.5"]

    _validate_outcome_constraints(outcome_constraints)


@pytest.mark.parametrize(
    "outcome_constraints, error_message",
    [
        (None, "outcome_constraints must be a list"),
        ([123], "Constraint at index 0 must be a string"),
        (
            ["TotalForceX > 10"],
            r"Constraint at index 0 \(TotalForceX > 10\) must contain either >= or <= operator",
        ),
        (
            ["TotalForceX <= 10 <= 20"],
            r"Constraint at index 0 \(TotalForceX <= 10 <= 20\) must be of form 'metric_name <= value'",
        ),
        (["<= 10"], r"Constraint at index 0 \(<= 10\) must have a metric name"),
        (
            ["TotalForceX <= abc"],
            r"Constraint at index 0 \(TotalForceX <= abc\): 'abc' must be a numeric value",
        ),
    ],
)
def test_validate_outcome_constraints_fails(outcome_constraints, error_message):
    with pytest.raises(
        expected_exception=InvalidArguments,
        match=error_message,
    ):
        _validate_outcome_constraints(outcome_constraints)


def test_validate_n_iters_success():
    n_iters = 10

    _validate_n_iters(n_iters)


@pytest.mark.parametrize(
    "n_iters, error_message",
    [
        (None, "n_iters must be an integer"),
        ("5", "n_iters must be an integer"),
        (0, "n_iters must be strictly positive"),
        (-5, "n_iters must be strictly positive"),
    ],
)
def test_validate_n_iters_fails(n_iters, error_message):
    with pytest.raises(
        expected_exception=InvalidArguments,
        match=error_message,
    ):
        _validate_n_iters(n_iters)


def validate_max_displacement_success():
    max_displacement = [1, 0.25]
    bounding_boxes = [[0.1, 0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1.0, 1.1, 1.2]]

    _validate_max_displacement(max_displacement, bounding_boxes)

    _validate_max_displacement(None, bounding_boxes)


@pytest.mark.parametrize(
    "max_displacement, bounding_boxes, error_message",
    [
        (
            "not_a_list",
            [[0.1, 0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1.0, 1.1, 1.2]],
            "max_displacement must be a list",
        ),
        (
            123,
            [[0.1, 0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1.0, 1.1, 1.2]],
            "max_displacement must be a list",
        ),
        (
            {},
            [[0.1, 0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1.0, 1.1, 1.2]],
            "max_displacement must be a list",
        ),
        (
            ["this_is_a_string"],
            [[0.1, 0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1.0, 1.1, 1.2]],
            "max_displacement contains non-numeric values",
        ),
        (
            [1, "perhaps_a_string"],
            [[0.1, 0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1.0, 1.1, 1.2]],
            "max_displacement contains non-numeric values",
        ),
        (
            [None],
            [[0.1, 0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1.0, 1.1, 1.2]],
            "max_displacement contains non-numeric values",
        ),
        (
            [1.0],
            [[0.1, 0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1.0, 1.1, 1.2]],
            "Max displacement list and bounding boxes list must have the same number of items",
        ),
        (
            [1.0, 2.0, 3.0],
            [[0.1, 0.2, 0.3, 0.4, 0.5, 0.6]],
            "Max displacement list and bounding boxes list must have the same number of items",
        ),
        (
            [],
            [[0.1, 0.2, 0.3, 0.4, 0.5, 0.6], [0.7, 0.8, 0.9, 1.0, 1.1, 1.2]],
            "Max displacement list and bounding boxes list must have the same number of items",
        ),
    ],
)
def test_validate_max_displacement_fails(max_displacement, bounding_boxes, error_message):
    with pytest.raises(
        expected_exception=InvalidArguments,
        match=error_message,
    ):
        _validate_max_displacement(max_displacement, bounding_boxes)


def validate_axial_symmetry_success():
    _validate_axial_symmetry("x", None)
    _validate_axial_symmetry(None, None)


@pytest.mark.parametrize(
    "axial_symmetry, symmetries",
    [
        ("a", None),
        (5, None),
        ("not_axis", None),
        (-1, None),
        ("x", ["x", "y"]),
    ],
)
def test_validate_axial_symmetry_fails(axial_symmetry, symmetries):
    with pytest.raises(
        expected_exception=InvalidArguments,
    ):
        _validate_axial_symmetry(axial_symmetry, symmetries)


def test_run_non_parametric_optimization(simai_client, geometry_factory, httpx_mock):
    workspace_id = "insert_cool_reference"
    geometry = geometry_factory(workspace_id=workspace_id)

    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/workspaces/{workspace_id}/optimizations",
        status_code=202,
        json={"id": "wow", "state": "requested", "trial_runs": []},
    )
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/optimizations/wow/trial-runs",
        status_code=202,
        json={"id": "wow1", "state": "requested"},
    )
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/optimizations/wow/trial-runs",
        status_code=202,
        json={"id": "wow2", "state": "requested"},
    )
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/optimizations/wow/trial-runs",
        status_code=202,
        json={"id": "wow3", "state": "requested"},
    )
    threading.Timer(
        0.1,
        simai_client._api._handle_sse_event,
        args=[
            ServerSentEvent(
                data=json.dumps(
                    {
                        "type": "job",
                        "status": "successful",
                        "record": {
                            "id": "wow",
                            "state": "successful",
                            "initial_geometry_parameters": None,
                        },
                        "target": {"type": "optimization", "id": "wow"},
                    }
                )
            )
        ],
    ).start()
    for i in range(1, 4):
        threading.Timer(
            i / 10 + 0.1,
            simai_client._api._handle_sse_event,
            args=[
                ServerSentEvent(
                    data=json.dumps(
                        {
                            "type": "job",
                            "status": "successful",
                            "record": {
                                "id": f"wow{i}",
                                "state": "successful",
                                "is_feasible": "true",
                                "outcome_values": i,
                                "next_geometry_parameters": None,
                                "geometry": {
                                    "id": f"x{i}",
                                    "name": f"x{i}",
                                    "workspace_id": geometry._fields["workspace_id"],
                                },
                            },
                            "target": {"type": "optimization_trial_run", "id": f"wow{i}"},
                        }
                    )
                )
            ],
        ).start()
    results = simai_client.optimizations.run_non_parametric(
        geometry=geometry,
        bounding_boxes=[[0.1, 1, 0.1, 1, 0.1, 1]],
        symmetries=["x", "y", "z"],
        minimize=["TotalForceX"],
        scalars={"VelocityX": 10.5},
        n_iters=3,
    )
    assert len(results.list_objectives()) == 3
    assert results.list_objectives() == [1, 2, 3]
    assert len(results.list_geometries()) == 3
    assert results.list_geometries()[0].name == "x1"
    assert results.optimization.id == "wow"



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _ref_generate_random_geometries:

Generating Random Geometries
==========================================

This example demonstrates how to generate random geometries using random latent parameters.

Before you begin
-------------------------------------------

- Complete ":ref:`ref_build_model`" to train a Generative Design model.
- Ensure the model training completed successfully.

"""

###############################################################################
# Import necessary libraries
# -------------------------------------------

import os
import random

import ansys.simai.core as asc
from ansys.simai.core.data.geomai.predictions import GeomAIPrediction, GeomAIPredictionConfiguration

###############################################################################
# Configure your settings
# -------------------------------------------
# Update these variables with your specific settings:

ORGANIZATION = "my_organization"  # Replace with your organization name
PROJECT_NAME = "new-bracket-project"  # Replace with your project name
WORKSPACE_NAME = "new-bracket-project #1"  # Typically "{PROJECT_NAME} #{number}"
OUTPUT_DIR = "random_geometries"  # Directory to save generated geometries
NUM_GEOMETRIES = 5  # Number of random geometries to generate
RESOLUTION = (100, 100, 100)  # Output resolution (x, y, z)

###############################################################################
# Initialize the client and get the workspace
# -------------------------------------------
# Connect to the instance:

simai_client = asc.SimAIClient(organization=ORGANIZATION)
geomai_client = simai_client.geomai

###############################################################################
# Retrieve the trained workspace by its name:

workspace = geomai_client.workspaces.get(name=WORKSPACE_NAME)
print(f"Using workspace: {workspace.name}")


###############################################################################
# Get the number of latent parameters
# -------------------------------------------
# The number of latent parameters is defined during model training:

latent_dict = workspace.get_latent_parameters()

nb_latent_params = len(next(iter(latent_dict.values())))
print(f"Workspace uses {nb_latent_params} latent parameters")

###############################################################################
# Create output directory
# -------------------------------------------
# Create a directory to save the generated geometries:

output_dir = os.path.join(OUTPUT_DIR, workspace.name)
os.makedirs(output_dir, exist_ok=True)
print(f"Output directory: {output_dir}")

###############################################################################
# Generate random geometries
# -------------------------------------------
# Generate geometries by creating random latent parameter vectors.
# Each latent parameter is randomly sampled from a standard normal distribution.

predictions: list[GeomAIPrediction] = []

print(f"\nGenerating {NUM_GEOMETRIES} random geometries...")
for i in range(NUM_GEOMETRIES):
    # Generate random latent parameters (from standard normal distribution)
    latent_params = [random.gauss(0, 1) for _ in range(nb_latent_params)]

    # Create prediction configuration
    config = GeomAIPredictionConfiguration(
        latent_params=latent_params,
        resolution=RESOLUTION,
    )

    # Run the prediction
    prediction = geomai_client.predictions.run(config, workspace)
    print(f"Prediction {i + 1}/{NUM_GEOMETRIES}: {prediction.id} started...")
    predictions.append(prediction)


###############################################################################
# Download generated geometries
# -------------------------------------------

for i, prediction in enumerate(predictions):
    # Wait for prediction to complete
    if prediction.wait(timeout=600):  # Wait up to 10 minutes
        if prediction.has_failed:
            print(f"✗ Prediction {i + 1} failed: {prediction.failure_reason}")
            continue

        # Download the generated geometry
        output_path = os.path.join(output_dir, f"random_{i + 1:02d}_{prediction.id}.vtp")
        geomai_client.predictions.download(prediction.id, output_path)
        print(f"✓ Saved geometry to {output_path}")
    else:
        print(f"✗ Prediction {i + 1} timed out")

###############################################################################
# The downloaded VTP files can be used for:
#
# - Visualization in your usual solver.
# - SimAI training data or predictions.
# - Further analysis and post-processing.

###############################################################################
# Tips for better results
# -------------------------------------------
#
# - Latent parameters typically range from -3 to +3 for meaningful results.
# - Adjust the resolution to balance quality and file size.
# - Increase timeout for complex geometries.
# - Use the workspace's latent space statistics (min, max) for better sampling.

###############################################################################
# Next steps
# -------------------------------------------
# To generate geometries with more control, see :ref:`ref_interpolate_geometries`.



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import platform
from pathlib import Path

import pytest

import ansys.simai.core.errors as err
from ansys.simai.core.utils.config_file import _scan_defaults_config_paths, get_config


@pytest.mark.skipif(platform.system() == "Windows", reason="XDG_CONFIG_HOME is for unix systems")
def test_xdg_config_home_is_respected(monkeypatch, tmpdir):
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmpdir))
    config_file_path = Path(tmpdir) / "ansys_simai.conf"
    config_file_path.touch()
    assert _scan_defaults_config_paths() == config_file_path


def test_get_config_invalid_path():
    with pytest.raises(err.ConfigurationNotFoundError):
        get_config(path="/")


def test_get_config_invalid_profile(tmpdir):
    path_config = Path(tmpdir) / "config"
    with open(path_config, "w+") as f:
        f.write("[default]\n")
        f.flush()
        with pytest.raises(err.InvalidConfigurationError):
            get_config(profile="kaboom", path=path_config)


def test_get_config_ignore_missing(mocker):
    mocker.patch(
        "ansys.simai.core.utils.config_file._scan_defaults_config_paths",
        return_value=None,
    )
    config = get_config(ignore_missing=True, organization="kangarooooo")
    assert config == {"organization": "kangarooooo"}



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# ruff: noqa: S311
import random
import threading
import time
from collections.abc import Callable
from contextlib import contextmanager

import httpx
import httpx_retries
import pytest

from ansys.simai.core import SimAIClient
from ansys.simai.core.api.client import ApiClient
from ansys.simai.core.data.geometries import Geometry, GeometryDirectory
from ansys.simai.core.data.models import Model
from ansys.simai.core.data.post_processings import (
    PostProcessing,
    PostProcessingDirectory,
)
from ansys.simai.core.data.predictions import Prediction
from ansys.simai.core.data.projects import Project
from ansys.simai.core.data.training_data import TrainingData
from ansys.simai.core.data.workspaces import Workspace
from ansys.simai.core.utils.configuration import ClientConfig


@pytest.fixture(scope="session")
def api_client():
    """A ApiClient object with bogus URL"""
    clt = ApiClient(
        config=ClientConfig(
            url="https://test.test",
            _disable_authentication=True,
            no_sse_connection=True,
            organization="ExtraClaquette",
        )
    )
    yield clt


# warning: even with scope=function, the same SimAIClient is kept,
# thus previous test data exist in different methods
@pytest.fixture(scope="function")
def simai_client():
    client = SimAIClient(
        url="https://test.test",
        _disable_authentication=True,
        no_sse_connection=True,
        skip_version_check=True,
        organization="ExtraCorp",
    )
    client._current_workspace = Workspace(
        client,
        client._workspace_directory,
        {"id": "whatever", "name": "not-an-actual-workspace-please-set-it-if-needed"},
    )
    yield client


@pytest.fixture(scope="function")
def sse_mixin(simai_client):
    yield ApiClient(
        simai_client=simai_client,
        config=ClientConfig(
            url="https://test.test",
            _disable_authentication=True,
            no_sse_connection=True,
            organization="PhantomOrg",
        ),
    )


@pytest.fixture(scope="function")
def post_processing_factory(simai_client) -> PostProcessing:
    """Returns a function to create a :py:class:`PostProcessing`."""

    def _factory(prediction=None, **kwargs) -> PostProcessing:
        kwargs.setdefault("id", str(random.random()))
        if prediction and "prediction_id" not in kwargs:
            kwargs["prediction_id"] = prediction.id
        kwargs.setdefault("state", "successful")
        return simai_client._post_processing_directory._model_from(kwargs, prediction=prediction)

    return _factory


@pytest.fixture(scope="function")
def prediction_factory(simai_client) -> Prediction:
    """Returns a function to create a :py:class:`Prediction`."""

    def _factory(post_processings=None, geometry=None, **kwargs) -> Prediction:
        if "id" not in kwargs:
            if "boundary_conditions" in kwargs:
                kwargs["id"] = "pred-" + "-".join(
                    str(s) for s in kwargs["boundary_conditions"].values()
                )
            else:
                kwargs["id"] = str(random.random())
        if geometry and "geometry_id" not in kwargs:
            kwargs["geometry_id"] = geometry.id
        kwargs.setdefault("boundary_conditions", {"Vx": 10.01, "Vy": 0.0009})
        kwargs.setdefault("state", "successful")
        kwargs.setdefault("confidence_score", None)
        kwargs.setdefault("raw_confidence_score", None)
        prediction = simai_client._prediction_directory._model_from(kwargs)
        if post_processings is not None:
            # If we passed post-processings as parameter,
            # link them with the prediction.
            # Evidently PP should not be passed directly if the aim
            # is to test creation requests.
            prediction.get_post_processings = lambda: post_processings
            for pp in post_processings:
                pp._prediction = prediction

                # add post processing to prediction
                pred_pp = prediction._post_processings
                params = pp.fields.get("location", {})
                params_frozen = frozenset(params.items())
                pp_type = PostProcessingDirectory._data_model_for_type_name(pp._api_name())
                if pp_type not in pred_pp._post_processings:
                    pred_pp._post_processings[pp_type] = {}
                pred_pp._post_processings[pp_type][params_frozen] = pp
        return prediction

    return _factory


@pytest.fixture()
def geometry_factory(simai_client) -> Geometry:
    """Returns a function to create a :py:class:`Geometry`."""

    def _factory(predictions=None, **kwargs) -> Geometry:
        kwargs.setdefault("id", str(random.random()))
        kwargs.setdefault("name", kwargs["id"])
        kwargs.setdefault("state", "successful")
        if not predictions:
            predictions = []
        kwargs["predictions"] = [s.id for s in predictions]

        geometry = simai_client._geometry_directory._model_from(kwargs)
        geometry.get_predictions = lambda: predictions
        return geometry

    return _factory


@pytest.fixture()
def training_data_factory(simai_client) -> TrainingData:
    """Returns a function to create a :py:class:`TrainingData`."""

    def _factory(**kwargs) -> TrainingData:
        kwargs.setdefault("id", str(random.random()))
        kwargs.setdefault("name", kwargs["id"])
        kwargs.setdefault("state", "successful")

        training_data = simai_client._training_data_directory._model_from(kwargs)
        return training_data

    return _factory


@pytest.fixture()
def project_factory(simai_client) -> Project:
    """Returns a function to create a :py:class:`Project`."""

    def _factory(**kwargs) -> Project:
        kwargs.setdefault("id", str(random.random()))
        kwargs.setdefault("name", kwargs["id"])

        project = simai_client._project_directory._model_from(kwargs)
        return project

    return _factory


@pytest.fixture()
def geometry_directory(simai_client):
    yield GeometryDirectory(client=simai_client)


@pytest.fixture()
def create_mock_geometry():
    """Creates a lightweight geometry, not tracked by simai_client"""

    def _factory(id, predictions=None, **kwargs):
        geometry = Geometry(None, None, {"id": id, "name": str(id), "metadata": kwargs})
        geometry.get_predictions = lambda: predictions or []
        return geometry

    return _factory


@pytest.fixture(scope="function")
def global_coefficient_request_factory(simai_client):
    """Returns a function to create a global coefficient request object."""

    def _factory(**kwargs):
        return simai_client._process_gc_formula_directory._model_from(**kwargs)

    return _factory


@pytest.fixture()
def model_factory(simai_client) -> Model:
    """Returns a function to create a :py:class:`Model`."""

    def _factory(**kwargs) -> Model:
        kwargs.setdefault("id", str(random.random()))
        kwargs.setdefault("name", kwargs["id"])

        model = simai_client._model_directory._model_from(kwargs)
        return model

    return _factory


@pytest.fixture
def delayed_events():
    """
    Fixture to test Event.wait() with sequential events.

    This fixture provides a manager for scheduling events to happen at specific times,
    allowing for controlled testing of wait() behavior.

    Usage:
        def test_something(delayed_events):
            # Schedule events to occur
            delayed_events.add(lambda: model.update_state("processing"))
            delayed_events.add(lambda: model.complete())

            # Start the background thread that will trigger events
            delayed_events.start()

            # Call the wait method you're testing
            result = model.wait()

            # Make assertions
            assert result is True
    """

    class DelayedEventManager:
        def __init__(self) -> None:
            self.events = []
            self.thread = None

        def add(self, callback: Callable[[], None], delay: float = 0.1):
            """Add a callback to be executed after the specified delay (in seconds)"""
            self.events.append((delay, callback))

        def start(self):
            """Start the thread that will execute all scheduled events"""

            def run_events():
                start_time = time.time()

                for delay, callback in self.events:
                    time_to_wait = delay - (time.time() - start_time)
                    if time_to_wait > 0:
                        time.sleep(time_to_wait)
                    callback()

            self.thread = threading.Thread(target=run_events)
            self.thread.daemon = True
            self.thread.start()
            return self

        def join(self):
            """Wait for all events to complete"""
            if self.thread:
                self.thread.join()

    manager = DelayedEventManager()

    yield manager

    manager.join()


@contextmanager
def disable_http_retry(clt: ApiClient, url: str):
    "Avoids slow tests due to retry backoff"
    transport = clt._session._transport_for_url(httpx.URL(url))
    original_retry = transport.retry
    transport.retry = httpx_retries.Retry(total=0)
    try:
        yield
    finally:
        transport.retry = original_retry



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import urllib.parse
from io import BytesIO
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.simai.core.data.geometries import Geometry


def test_geometries_list_no_parameter(simai_client, httpx_mock):
    expected_params = urllib.parse.urlencode(
        {
            "filters": {},
            "workspace": simai_client._current_workspace.id,
        }
    )
    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/geometries/?{expected_params}",
        json={},
        status_code=200,
    )
    list(simai_client.geometries.list())


def test_geometries_filter(simai_client, httpx_mock):
    expected_params = urllib.parse.urlencode(
        {
            "filters": json.dumps(
                {
                    "DIAMETER": 12.5,
                    "SAUCE": "cream",
                }
            ),
            "workspace": simai_client._current_workspace.id,
        }
    )
    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/geometries/?{expected_params}",
        json={},
        status_code=200,
    )
    list(simai_client.geometries.list(filters={"DIAMETER": 12.5, "SAUCE": "cream"}))


def test_geometries_run_prediction(geometry_factory, httpx_mock):
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/geometries/geom-0/predictions",
        json={"id": "pred-0"},
        status_code=200,
    )
    geometry = geometry_factory(id="geom-0")
    geometry.run_prediction(Vx=10.5)


def test_geometries_run_prediction_dict_bc(geometry_factory, httpx_mock):
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/geometries/geom-0/predictions",
        json={"id": "pred-0"},
        status_code=200,
    )
    geometry = geometry_factory(id="geom-0")
    geometry.run_prediction({"Vx": 10.5})


def test_geometries_run_prediction_no_bc(geometry_factory, httpx_mock):
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/geometries/geom-0/predictions",
        json={"id": "pred-0"},
        status_code=200,
    )
    geometry = geometry_factory(id="geom-0")
    geometry.run_prediction()


def test_geometries_upload_point_cloud(geometry_factory, httpx_mock):
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/geometries/geom-0/point-cloud",
        json={"point_cloud": {"id": "pc-0"}, "upload_id": "123"},
        status_code=200,
    )
    httpx_mock.add_response(
        method="PUT",
        url="https://test.test/point-clouds/pc-0/part",
        json={"url": "https://s3.test/pc-0/part"},
        status_code=200,
    )
    httpx_mock.add_response(
        method="PUT", url="https://s3.test/pc-0/part", headers={"ETag": "SaladeTomateOignon"}
    )
    httpx_mock.add_response(
        method="POST", url="https://test.test/point-clouds/pc-0/complete", status_code=204
    )

    geometry = geometry_factory(id="geom-0")
    file = BytesIO(b"Hello World")
    geometry.upload_point_cloud((file, "my-point-cloud.vtp"))
    assert geometry.point_cloud == {"id": "pc-0"}


def test_geometries_delete_point_cloud(geometry_factory, httpx_mock):
    httpx_mock.add_response(
        method="DELETE", url="https://test.test/point-clouds/point-cloud-0", status_code=204
    )

    geometry = geometry_factory(id="geom-0", point_cloud={"id": "point-cloud-0"})
    geometry.delete_point_cloud()
    assert geometry.point_cloud is None


def test_geometries_delete_point_cloud_cleares_pp_cache(
    geometry_factory, prediction_factory, post_processing_factory, httpx_mock
):
    httpx_mock.add_response(
        method="DELETE", url="https://test.test/point-clouds/point-cloud-0", status_code=204
    )

    custom_volume_point_cloud = post_processing_factory(
        type="CustomVolumePointCloud", prediction_id="prediction-0"
    )
    prediction = prediction_factory(
        id="prediction-0", post_processings=[custom_volume_point_cloud], geometry_id="geom-0"
    )
    geometry = geometry_factory(
        id="geom-0", predictions=[prediction], point_cloud={"id": "point-cloud-0"}
    )

    assert custom_volume_point_cloud in prediction.post._local_post_processings
    geometry.delete_point_cloud()
    assert custom_volume_point_cloud not in prediction.post._local_post_processings


def test_geometries_rename(simai_client, httpx_mock):
    geometry: Geometry = simai_client._geometry_directory._model_from(
        {"id": "0011", "name": "riri"}
    )

    httpx_mock.add_response(
        method="PATCH",
        url="https://test.test/geometries/0011",
        status_code=204,
    )
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/geometries/0011",
        json={"id": "0011", "name": "fifi"},
        status_code=200,
    )

    geometry.rename("fifi")
    assert geometry.name == "fifi"


def test_geometries_list_predictions(geometry_factory, prediction_factory, httpx_mock):
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/geometries/geom-0/predictions",
        json=[
            {"id": "pred-0"},
            {"id": "pred-1"},
        ],
        status_code=200,
    )
    geometry = geometry_factory(id="geom-0")
    predictions = geometry.list_predictions()
    assert len(predictions) == 2
    assert predictions[0].id == "pred-0"
    assert predictions[1].id == "pred-1"



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import httpcore
import httpx
import pytest
from pydantic import HttpUrl
from pytest_httpx import HTTPXMock

from ansys.simai.core import SimAIClient
from ansys.simai.core.errors import ApiClientError, NotFoundError

from .conftest import disable_http_retry


def test_construct_default_url():
    """WHEN No URL is passed to SimAIClient
    THEN A default one is used
    """
    client = SimAIClient(
        no_sse_connection=True,
        _disable_authentication=True,
        skip_version_check=True,
        organization="ExtraBanane",
    )
    assert client._api._url_prefix == HttpUrl("https://api.simai.ansys.com/v2/")


def test_retry_on_5XX(api_client, httpx_mock: HTTPXMock):
    url = "https://try.me/"
    httpx_mock.add_response(method="GET", url=url, status_code=503)
    httpx_mock.add_response(method="GET", url=url, status_code=504)
    httpx_mock.add_response(method="GET", url=url, status_code=200, json={"wat": "hyperdrama"})
    resp = api_client._get(url)
    assert resp == {"wat": "hyperdrama"}


def test_404_response_raises_not_found_error(api_client, httpx_mock: HTTPXMock):
    """WHEN ApiClient gets a 404 response
    THEN a NotFoundError is raised
    """
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/expected_format",
        json={"message": "beep-boop"},
        status_code=404,
    )
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/only_status_format",
        json={"status": "not found"},
        status_code=404,
    )
    httpx_mock.add_response(method="GET", url="https://test.test/no_content", status_code=404)

    with pytest.raises(NotFoundError) as exc_info:
        api_client._get("expected_format")
    assert exc_info.value.args[0] == "beep-boop"
    with pytest.raises(NotFoundError) as exc_info:
        api_client._get("only_status_format")
    assert exc_info.value.args[0] == "not found"
    with pytest.raises(NotFoundError) as exc_info:
        api_client._get("no_content")
    assert exc_info.value.args[0] == "Not Found"


@pytest.mark.parametrize(
    "code, reason",
    [
        (400, "Bad Request"),
        (401, "Unauthorized"),
        (403, "Forbidden"),
        # (404, "Not Found"), # raises NotFoundError instead # noqa: ERA001
        (405, "Method Not Allowed"),
        (500, "Internal Server Error"),
        (501, "Not Implemented"),
        (502, "Bad Gateway"),
    ],
)
def test_4XX_5XX_responses_raise_api_client_error_no_json(
    api_client, httpx_mock: HTTPXMock, code, reason
):
    """WHEN ApiClient gets a 4XX or 5XX response without details in json
    THEN a ApiClientError is raised with the code and reason
    """
    httpx_mock.add_response(method="GET", url=f"https://test.test/{code}", status_code=code)
    with disable_http_retry(api_client, "https://test.test/"):
        with pytest.raises(ApiClientError) as exc_info:
            api_client._get(f"{code}")
        assert exc_info.value.args[0] == f"{code} {reason}"


@pytest.mark.parametrize(
    "json, message",
    [
        ({"errors": {"beep": "boop"}}, "{'beep': 'boop'}"),
        ({"message": "tomato"}, "tomato"),
        (
            {"status": "I'm sorry Dave, I'm afraid I can't do that"},
            "I'm sorry Dave, I'm afraid I can't do that",
        ),
        (None, "Bad Request"),
    ],
)
def test_4XX_5XX_responses_raise_api_client_error_with_json(api_client, httpx_mock, json, message):
    """WHEN ApiClient gets a 4XX or 5XX response without details in json
    THEN a ApiClientError is raised with the code and message from the json or fallback
    """
    httpx_mock.add_response(
        method="GET", url="https://test.test/errors", json=json, status_code=400
    )

    with pytest.raises(ApiClientError) as exc_info:
        api_client._get("errors")
    assert exc_info.value.args[0] == f"400 {message}"


def test_use_system_proxies(mocker):
    """WHEN ApiClient is initialized on a system with default proxies
    THEN the system proxies are used
    """
    proxy_url = "https://bonzibuddy.org"
    mocker.patch(
        "httpx._utils.getproxies",
        return_value={"http": proxy_url},
    )

    client = SimAIClient(
        no_sse_connection=True,
        _disable_authentication=True,
        skip_version_check=True,
        organization="ExtraBanane",
    )
    transport = client._api._session._transport_for_url(httpx.URL("http://popo.org"))
    assert transport._sync_transport._pool._proxy_url == httpcore.URL(proxy_url)


def test_use_user_provided_proxies(mocker):
    """WHEN ApiClient is initialized with a specific https_proxy
    THEN the specific proxy is used (system proxies are ignored), the url is normalized
    """
    mocker.patch(
        "httpx._utils.getproxies",
        return_value={"https": "https://bonzibuddy.org"},
    )

    client = SimAIClient(
        no_sse_connection=True,
        _disable_authentication=True,
        skip_version_check=True,
        organization="ExtraBanane",
        https_proxy="https://😛.com",
    )
    transport = client._api._session._transport_for_url(httpx.URL("https://popo.org"))
    assert transport._sync_transport._pool._proxy_url == httpcore.URL("https://xn--528h.com")



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _ref_create_project_upload_data:

Creating a GeomAI Project and Uploading Training Data
===========================================================

This example demonstrates how to connect to the instance, create a new project, and upload geometry files as training data.

Before you begin
-------------------------------------------

Make sure you have:

- Valid SimAI credentials and organization access.
- A folder containing geometry files (.vtp or .stl format).
- The ``ansys-simai-core`` library installed.

"""

###############################################################################
# Import necessary libraries
# -------------------------------------------

import os

import ansys.simai.core as asc
from ansys.simai.core.errors import NotFoundError

###############################################################################
# Configure your settings
# -------------------------------------------
# Update these variables with your specific settings:

ORGANIZATION = "my_organization"  # Replace with your organization name
DATASET_PATH = "path/to/your/data/folder"  # Folder with .vtp or .stl files
PROJECT_NAME = "new-bracket-project"  # Your project name

###############################################################################
# Create the client
# -------------------------------------------
# Create a client to use the PySimAI library. This client will be the
# entrypoint for all Generative Design objects.

simai_client = asc.SimAIClient(organization=ORGANIZATION)
geomai_client = simai_client.geomai
available_tds = geomai_client.training_data.list()

###############################################################################
# Create or retrieve a project
# -------------------------------------------
# Try to get an existing project by name, or create it if it does not exist:

try:
    project = geomai_client.projects.get(name=PROJECT_NAME)
    print(f"Using existing project: {PROJECT_NAME}")
except NotFoundError:
    project = geomai_client.projects.create(PROJECT_NAME)
    print(f"Created new project: {PROJECT_NAME}")

print(f"Current project: {project.name}")

###############################################################################
# Upload training data to the project
# -------------------------------------------
# Loop through all geometry files in your dataset folder and upload them.
# The script handles duplicates by checking if the data already exist.

print("\nUploading training data files:")
successful_uploads = 0
failed_uploads = 0

for fname in os.listdir(DATASET_PATH):
    td_name = os.path.splitext(fname)[0]
    fpath = os.path.join(DATASET_PATH, fname)

    # Skip non-geometry files
    if not fname.lower().endswith((".vtp", ".stl")):
        print(f"Skipping non-geometry file: {fname}")
        continue

    # Check if training data already exist
    existing_tds = [td for td in available_tds if td.name == td_name]
    if existing_tds:
        print(f"Training data '{fname}' already exists in the datalake. Skipping upload.")
        td = existing_tds[0]
        try:
            td.add_to_project(project)
            print(f"✓ Added existing '{fname}' to project '{project.name}'")
            successful_uploads += 1
        except Exception as e:
            print(f"✗ Failed to add existing '{fname}' to project: {e}")
            failed_uploads += 1
        continue

    # Upload new training data
    try:
        training_data = geomai_client.training_data.create_from_file(file=fpath, project=project)
        print(f"✓ Uploaded '{fname}' -> ID: {training_data.id}")
        successful_uploads += 1
    except Exception as e:
        print(f"✗ Failed to upload '{fname}': {e}")
        failed_uploads += 1

print(f"\nUpload summary: {successful_uploads} successful, {failed_uploads} failed")

###############################################################################
# Check and wait for data processing
# -------------------------------------------
# After uploading, the instance needs to process the geometries. This script
# displays the progress of the data processing.

project_data = project.data()


print("\nWaiting for data processing to complete...")
for data in project_data:
    print(f"Processing '{data.name}'...")
    data.wait()
    if data.is_ready:
        print(f"✓ Data '{data.name}' is ready")
    else:
        print(f"✗ Data '{data.name}' failed: {data.failure_reason}")

###############################################################################
# Display project status summary
# -------------------------------------------
# Show a summary of the project's data processing status:

project_data = project.data()

print("\nProject Summary")
print("=" * 50)
ready_data = [data for data in project_data if data.is_ready]
not_ready_data = [data for data in project_data if not data.is_ready]

print(f"Total data in project: {len(project_data)}")
print(f"Ready data: {len(ready_data)} of {len(project_data)}")
print(f"Not ready data: {len(not_ready_data)} of {len(project_data)}")

if not_ready_data:
    print(
        "\nFailed data processing details:\n"
        "Having an 'invalid geometry' means that the geometry is not compatible with Generative Design. "
        "Please check the geometry file for errors or issues (watertightness and manifold).\n"
    )
    for data in not_ready_data:
        print(f"- {data.name}: {data.failure_reason}")

###############################################################################
# Next steps
# -------------------------------------------
# Once all data is ready, you can proceed to build a model.
# See the next example: :ref:`ref_build_model`.



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest
from pydantic_core import ValidationError

from ansys.simai.core.utils.configuration import ClientConfig


@pytest.mark.parametrize(
    "inputs,password,config,expected_output",
    [
        (
            [
                "12_monkeys",
                "timmy",
                "12345",
            ],  # list of simulated user inputs (other than password)
            "shakaponk",  # password for getpass if needed
            {
                "credentials": {"totp_enabled": True}
            },  # the configuration the method will be called with
            {
                "organization": "12_monkeys",
                "credentials": {
                    "username": "timmy",
                    "password": "shakaponk",
                    "totp": "12345",
                },
            },  # what the method should return
        ),
        (
            ["12345"],
            None,
            {
                "credentials": {
                    "totp_enabled": True,
                    "password": "shakaponk",
                    "username": "timmy",
                },
                "organization": "12_monkeys",
            },
            {
                "credentials": {
                    "username": "timmy",
                    "password": "shakaponk",
                    "totp": "12345",
                },
                "organization": "12_monkeys",
            },
        ),
        (
            [],
            None,
            {
                "credentials": {
                    "username": "timmy",
                    "password": "shakaponk",
                },
                "organization": "12_monkeys",
            },
            {
                "credentials": {"username": "timmy", "password": "shakaponk"},
                "organization": "12_monkeys",
            },
        ),
        (
            ["timmy"],
            "shakaponk",
            {"organization": "12_monkeys", "credentials": {}},
            {
                "credentials": {
                    "username": "timmy",
                    "password": "shakaponk",
                },
                "organization": "12_monkeys",
            },
        ),
        (
            [],
            None,
            {"organization": "12_monkeys"},
            {"organization": "12_monkeys"},
        ),
        (
            [],
            None,
            {  # interactive mode is OFF, and password is missing from credentials
                "organization": "12_monkeys",
                "interactive": False,
                "credentials": {
                    "username": "timmy",
                },
            },
            {"expect_error": True, "type": ValidationError, "error_count": 1},
        ),
        (
            [],
            None,
            {  # interactive mode is OFF, and organization is missing
                "interactive": False,
                "credentials": {"username": "timmy", "password": "teas"},
            },
            {"expect_error": True, "type": ValidationError, "error_count": 1},
        ),
        (
            [],
            None,
            {  # interactive mode is OFF, and organization and password are missing
                "interactive": False,
                "credentials": {"username": "timmy"},
            },
            {"expect_error": True, "type": ValidationError, "error_count": 2},
        ),
        (
            [],
            None,
            {  # interactive mode is OFF, and organization and password are missing, and url is not valid
                "url": "123",
                "interactive": False,
                "credentials": {"username": "timmy"},
            },
            {"expect_error": True, "type": ValidationError, "error_count": 3},
        ),
        (
            [],
            None,
            {  # interactive mode is ON, and credentials are missing
                "organization": "12_monkeys",
                "interactive": False,
            },
            {"expect_error": True, "type": ValidationError, "error_count": 1},
        ),
        (
            [],
            "pass",
            {  # sanity check; interactive is set explicitly to ON
                "interactive": True,
                "credentials": {"username": "timmy"},
                "organization": "12_monkeys",
            },
            {
                "credentials": {
                    "username": "timmy",
                    "password": "pass",
                },
                "organization": "12_monkeys",
            },
        ),
    ],
)
def test_get_authentication_configuration(inputs, password, config, expected_output, mocker):
    """GIVEN A list of prompt inputs, a configuration and an expected output
    WHEN getting authentication info for the config
    THEN the expected output is returned
    """
    mocker.patch("builtins.input", side_effect=inputs)
    mocker.patch("getpass.getpass", return_value=password)
    if expected_output.get("expect_error", None):
        # if an error is expected, catch the exception type and assert the error count
        with pytest.raises(expected_output.get("type")) as ex:
            client_config = ClientConfig(**config)
        assert ex.value.error_count() == expected_output.get("error_count")
    else:
        client_config = ClientConfig(**config).model_dump(exclude_none=True)
        # the config contains many fields, here we only test a subset
        tested_fields = ["organization", "credentials"]
        for f in list(client_config.keys()):
            if f not in tested_fields:
                client_config.pop(f)
        assert client_config == expected_output



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import TYPE_CHECKING

from ansys.simai.core.data.geomai.models import GeomAIModelConfiguration

if TYPE_CHECKING:
    from ansys.simai.core.data.geomai.workspaces import GeomAIWorkspace


NB_EPOCHS = 50
NB_LATENT_PARAMS = 10

MODEL_CONF_RAW = {"nb_epochs": NB_EPOCHS, "nb_latent_param": NB_LATENT_PARAMS, "build_preset": None}


def test_geomai_workspace_download_mer(simai_client, httpx_mock):
    """WHEN downloading mer zip file
    THEN the content of the file matches the content of the response.
    """

    workspace: GeomAIWorkspace = simai_client.geomai._workspace_directory._model_from(
        {"id": "abc123", "name": "HL3"}
    )
    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/geomai/workspaces/{workspace.id}/model-evaluation-report",
        text="mer-geomai",
        status_code=200,
    )

    in_memory = workspace.download_model_evaluation_report()
    data_in_file = in_memory.readline()
    assert data_in_file.decode("ascii") == "mer-geomai"


def test_geomai_workspace_rename(simai_client, httpx_mock):
    workspace: GeomAIWorkspace = simai_client.geomai._workspace_directory._model_from(
        {"id": "0011", "name": "riri"}
    )

    httpx_mock.add_response(
        method="PATCH",
        url="https://test.test/geomai/workspaces/0011",
        status_code=204,
    )
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/geomai/workspaces/0011",
        json={"id": "0011", "name": "fifi"},
        status_code=200,
    )

    workspace.rename("fifi")
    assert workspace.name == "fifi"


def test_geomai_workspace_get_latent_parameters(simai_client, httpx_mock):
    workspace: GeomAIWorkspace = simai_client.geomai._workspace_directory._model_from(
        {"id": "abc123", "name": "HL3"}
    )
    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/geomai/workspaces/{workspace.id}/model/latent-parameters-json",
        text='{"geometry1": [1,2,3]}',
        status_code=200,
    )

    latent_parameters = workspace.get_latent_parameters()
    assert isinstance(latent_parameters, dict)
    assert latent_parameters == {"geometry1": [1, 2, 3]}


def test_get_workspace_model_configuration(mocker, simai_client, httpx_mock, training_data_factory):
    workspace: GeomAIWorkspace = simai_client.geomai._workspace_directory._model_from(
        {"id": "0011", "name": "riri"}
    )

    httpx_mock.add_response(
        method="GET",
        url="https://test.test/workspaces/0011/model/configuration",
        json=MODEL_CONF_RAW,
        status_code=200,
    )

    assert workspace.model_configuration.model_dump() == MODEL_CONF_RAW
    assert isinstance(workspace.model_configuration, GeomAIModelConfiguration)


def test_geomai_workspace_list_predictions(simai_client, httpx_mock):
    workspace: GeomAIWorkspace = simai_client.geomai._workspace_directory._model_from(
        {"id": "abc123", "name": "HL3"}
    )
    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/geomai/workspaces/{workspace.id}/predictions",
        json=[
            {"id": "pred1", "name": "Prediction 1"},
            {"id": "pred2", "name": "Prediction 2"},
        ],
        status_code=200,
    )

    predictions = workspace.list_predictions()
    assert len(predictions) == 2
    assert predictions[0].id == "pred1"
    assert predictions[1].id == "pred2"



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, List, Optional

from ansys.simai.core.utils.auth import get_offline_token

if TYPE_CHECKING:
    from ansys.simai.core.client import SimAIClient


@dataclass
class Consent:
    """Represents a consent granted by the current user to a client (like the SDK)."""

    client_id: str
    """The client ID that has been granted offline access."""
    created_date: datetime
    """When the consent was originally granted."""
    last_updated_date: datetime
    """When the consent was last updated."""
    granted_scopes: List[str]
    """List of scopes granted to this client."""

    def __repr__(self) -> str:
        return f"<Consent: {self.client_id}, created: {self.created_date!s}, last updated: {self.last_updated_date!s}>"


class ConsentDirectory:
    """Provides access to consent management operations.

    Consents represent the authorization records that grant a client (like the SDK)
    permission to use offline tokens for authentication.

    Example:
        .. code-block:: python

            import ansys.simai.core as asc

            simai_client = asc.from_config()

            # List all consents
            consents = simai_client.me.consents.list()
            for consent in consents:
                print(f"Client: {consent.client_id}, Created: {consent.created_date}")

            # Revoke a specific consent
            simai_client.me.consents.revoke("sdk")
    """

    def __init__(self, client: "SimAIClient"):
        self._client = client

    def list(self) -> List[Consent]:
        """List all offline access consents for the current user.

        Returns:
            List of :class:`Consent` objects representing the user's offline access consents.

        Example:
            .. code-block:: python

                consents = simai_client.me.consents.list()
                for consent in consents:
                    print(f"Client: {consent.client_id}")
                    print(f"  Created: {consent.created_date}")
                    print(f"  Scopes: {consent.granted_scopes}")
        """
        raw_consents = self._client._api.get_offline_tokens()
        return [
            Consent(
                client_id=consent["client_id"],
                created_date=datetime.fromisoformat(consent["created_date"]),
                last_updated_date=datetime.fromisoformat(consent["last_updated_date"]),
                granted_scopes=consent.get("granted_scopes", []),
            )
            for consent in raw_consents
        ]

    def revoke(self, client_id: str) -> None:
        """Revoke an offline access consent for the current user.

        This will invalidate the consent associated with the given client ID,
        preventing any further use of offline tokens for that client.

        Args:
            client_id: The client ID of the consent to revoke.

        Raises:
            NotFoundError: If no consent exists for the given client ID.

        Example:
            .. code-block:: python

                # Revoke consent for the SDK client
                simai_client.me.consents.revoke("sdk")

        Warning:
            Revoking a consent will invalidate any offline tokens associated with
            that client. Server-side workflows using those tokens will stop working.
        """
        self._client._api.revoke_offline_token(client_id)


class CurrentUser:
    """Provides access to current user self-management operations.

    This class allows users to manage their own account settings, generate
    offline tokens, and manage consents.

    Example:
        .. code-block:: python

            import ansys.simai.core as asc

            simai_client = asc.from_config()

            # Generate an offline token
            token = simai_client.me.generate_offline_token()

            # List all consents
            consents = simai_client.me.consents.list()
            for consent in consents:
                print(f"Client: {consent.client_id}, Created: {consent.created_date}")

            # Revoke a specific consent
            simai_client.me.consents.revoke("sdk")
    """

    def __init__(self, client: "SimAIClient"):
        self._client = client
        self._consents = ConsentDirectory(client)

    @property
    def consents(self) -> ConsentDirectory:
        """Access consent management operations.

        Returns:
            :class:`ConsentDirectory` for managing offline access consents.
        """
        return self._consents

    def generate_offline_token(self) -> str:
        """Generate a new offline token for the current user.

        This creates a long-lived refresh token that can be used for authentication
        without requiring interactive login. The token is associated with the "sdk"
        client ID.

        Returns:
            The offline token string that can be stored and used for future authentication.

        Example:
            .. code-block:: python

                # Generate an offline token
                token = simai_client.me.generate_offline_token()
                print(f"Store this token securely: {token}")
        """
        config = self._client._config
        https_proxy: Optional[str] = str(config.https_proxy) if config.https_proxy else None
        tls_ca_bundle: Optional[str] = None
        if config.tls_ca_bundle is not None:
            if isinstance(config.tls_ca_bundle, str):
                tls_ca_bundle = config.tls_ca_bundle
            else:
                tls_ca_bundle = str(config.tls_ca_bundle)

        return get_offline_token(
            url=str(config.url),
            credentials=config.credentials,
            https_proxy=https_proxy,
            tls_ca_bundle=tls_ca_bundle,
        )



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


def test_post_processing_delete(simai_client, post_processing_factory, httpx_mock):
    """WHEN Calling delete() on a post-processing
    THEN the /delete endpoint is called
    ALSO the post-processing is not anymore registered in the Directory
    """
    httpx_mock.add_response(
        method="DELETE",
        url="https://test.test/post-processings/uninteresting-coeffs",
        status_code=204,
    )

    global_coefficients = post_processing_factory(
        id="uninteresting-coeffs", type="GlobalCoefficients"
    )
    global_coefficients.delete()
    assert "uninteresting-coeffs" not in simai_client._post_processing_directory._registry


def test_post_processing_global_coefficients_delete(prediction_factory, httpx_mock):
    """WHEN deleting a GlobalCoefficients post-processing from a prediction
    THEN there is a call to the DELETE endpoint
    ALSO a new call to pred.post.global_coefficients() re-runs the post-processing
    """
    pred = prediction_factory()
    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/predictions/{pred.id}/post-processings/GlobalCoefficients",
        json={
            "id": "pp-instance-one",
            "state": "successful",
            "location": {},
        },
        status_code=200,
    )
    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/predictions/{pred.id}/post-processings/GlobalCoefficients",
        json={
            "id": "pp-instance-two",
            "state": "successful",
            "location": {},
        },
        status_code=200,
    )
    httpx_mock.add_response(
        method="DELETE",
        url="https://test.test/post-processings/pp-instance-one",
        status_code=204,
    )

    global_coefficients = pred.post.global_coefficients()
    assert global_coefficients.id == "pp-instance-one"

    global_coefficients.delete()

    global_coefficients = pred.post.global_coefficients()
    assert global_coefficients.id == "pp-instance-two"


def test_post_processing_surface_evolution_delete(prediction_factory, httpx_mock):
    """WHEN deleting a SurfaceEvolution post-processing from a prediction
    THEN there is a call to the DELETE endpoint
    ALSO a new call to pred.post.surface_evolution() re-runs the post-processing
    """
    pred = prediction_factory()
    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/predictions/{pred.id}/post-processings/SurfaceEvol",
        json={
            "id": "im-the-first-one",
            "state": "successful",
            "type": "SurfaceEvol",
            "location": {"axis": "y", "delta": 9.5},
        },
        headers={},
        status_code=200,
    )
    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/predictions/{pred.id}/post-processings/SurfaceEvol",
        json={
            "id": "im-the-second-one",
            "state": "successful",
            "type": "SurfaceEvol",
            "location": {"axis": "y", "delta": 9.5},
        },
        headers={},
        status_code=200,
    )
    httpx_mock.add_response(
        method="DELETE",
        url="https://test.test/post-processings/im-the-first-one",
        status_code=204,
    )

    surface_evolution = pred.post.surface_evolution(axis="y", delta=9.5)
    assert surface_evolution.id == "im-the-first-one"

    surface_evolution.delete()

    surface_evolution = pred.post.surface_evolution(axis="y", delta=9.5)
    assert surface_evolution.id == "im-the-second-one"
